from collections.abc import Iterator
from inspect import isclass
from typing import Optional

import _pytest.logging
import pytest
from pytest_container import Container
from pytest_container.container import ContainerData

import tmt.log
import tmt.package_managers
import tmt.package_managers.apk
import tmt.package_managers.apt
import tmt.package_managers.dnf
import tmt.package_managers.rpm_ostree
import tmt.plugins
import tmt.steps.provision.podman
import tmt.utils
from tmt.package_managers import (
    FileSystemPath,
    Installable,
    Options,
    Package,
    PackageManager,
    PackageManagerClass,
    PackagePath,
    )
from tmt.steps.provision.podman import GuestContainer, PodmanGuestData
from tmt.utils import ShellScript

from . import MATCH, assert_log

# We will need a logger...
logger = tmt.Logger.create()
logger.add_console_handler()

# Explore available plugins
tmt.plugins.explore(logger)

# Local images created via `make images-tests`, reference to local registry
CONTAINER_FEDORA_RAWHIDE = Container(
    url='containers-storage:localhost/tmt/tests/container/fedora/rawhide/upstream:latest')
CONTAINER_FEDORA_41 = Container(
    url='containers-storage:localhost/tmt/tests/container/fedora/41/upstream:latest')
CONTAINER_FEDORA_40 = Container(
    url='containers-storage:localhost/tmt/tests/container/fedora/40/upstream:latest')
CONTAINER_FEDORA_39 = Container(
    url='containers-storage:localhost/tmt/tests/container/fedora/39/upstream:latest')
CONTAINER_CENTOS_STREAM_9 = Container(
    url='containers-storage:localhost/tmt/tests/container/centos/stream9/upstream:latest')
CONTAINER_CENTOS_7 = Container(
    url='containers-storage:localhost/tmt/tests/container/centos/7/upstream:latest')
CONTAINER_UBI_8 = Container(
    url='containers-storage:localhost/tmt/tests/container/ubi/8/upstream:latest')
CONTAINER_UBUNTU_2204 = Container(
    url='containers-storage:localhost/tmt/tests/container/ubuntu/22.04/upstream:latest')
CONTAINER_DEBIAN_127 = Container(
    url='containers-storage:localhost/tmt/tests/container/debian/12.7/upstream:latest')
CONTAINER_FEDORA_COREOS = Container(
    url='containers-storage:localhost/tmt/tests/container/fedora/coreos:stable')
CONTAINER_FEDORA_COREOS_OSTREE = Container(
    url='containers-storage:localhost/tmt/tests/container/fedora/coreos/ostree:stable')
CONTAINER_ALPINE = Container(
    url='containers-storage:localhost/tmt/tests/container/alpine:latest')

PACKAGE_MANAGER_DNF5 = tmt.package_managers._PACKAGE_MANAGER_PLUGIN_REGISTRY.get_plugin('dnf5')
PACKAGE_MANAGER_DNF = tmt.package_managers._PACKAGE_MANAGER_PLUGIN_REGISTRY.get_plugin('dnf')
PACKAGE_MANAGER_YUM = tmt.package_managers._PACKAGE_MANAGER_PLUGIN_REGISTRY.get_plugin('yum')
PACKAGE_MANAGER_APT = tmt.package_managers._PACKAGE_MANAGER_PLUGIN_REGISTRY.get_plugin('apt')
PACKAGE_MANAGER_RPMOSTREE = tmt.package_managers._PACKAGE_MANAGER_PLUGIN_REGISTRY \
    .get_plugin('rpm-ostree')
PACKAGE_MANAGER_APK = tmt.package_managers._PACKAGE_MANAGER_PLUGIN_REGISTRY.get_plugin('apk')


def has_legacy_dnf(container: ContainerData) -> bool:
    """
    Checks whether a container provides older ``dnf`` and ``yum``.

    At some point, Fedora switched to ``dnf5`` completely, and older
    ``dnf`` and ``yum`` commands are now mere symlinks to ``dnf5``.
    """

    if 'fedora' not in container.image_url_or_id \
            and 'centos' not in container.image_url_or_id:
        return False

    return container.image_url_or_id not in (
        CONTAINER_FEDORA_RAWHIDE.url,
        CONTAINER_FEDORA_41.url,
        CONTAINER_FEDORA_COREOS.url,
        CONTAINER_FEDORA_COREOS_OSTREE.url
        )


def has_dnf5_preinstalled(container: ContainerData) -> bool:
    """ Checks whether a container provides ``dnf5`` """

    return container.image_url_or_id in (
        CONTAINER_FEDORA_RAWHIDE.url,
        CONTAINER_FEDORA_41.url,
        CONTAINER_FEDORA_COREOS.url,
        CONTAINER_FEDORA_COREOS_OSTREE.url
        )


def assert_output(
        expected_output: Optional[str],
        stdout: Optional[str],
        stderr: Optional[str]) -> None:
    """
    Check that the expected output is present

    We don't care whether the expected string is in stdout or stderr.
    Just make sure the output is there.
    """

    # Nothing to do if there are no expectations
    if not expected_output:
        return

    combined_output = (stdout or "") + (stderr or "")
    assert combined_output != ""
    assert expected_output in combined_output


# Note: keep the list ordered by the most desired package manager to the
# least desired one. For most of the tests, the order is not important,
# but the list is used to generate the discovery tests as well, and the
# order is used to find out what package manager is expected to be
# discovered.
CONTAINER_BASE_MATRIX = [
    # Fedora
    (CONTAINER_FEDORA_RAWHIDE, PACKAGE_MANAGER_DNF5),

    (CONTAINER_FEDORA_41, PACKAGE_MANAGER_DNF5),

    (CONTAINER_FEDORA_40, PACKAGE_MANAGER_DNF5),
    (CONTAINER_FEDORA_40, PACKAGE_MANAGER_DNF),
    (CONTAINER_FEDORA_40, PACKAGE_MANAGER_YUM),

    # Supported in theory, but in practice, dnf5 is not going to
    # officially be part of Fedora 39, and we are seeing some strange
    # and broken behavior ruining our tests. On top of that, Fedora 39
    # is going to be EOL very soon, therefore disabling this particular
    # combination.
    # (CONTAINER_FEDORA_39, PACKAGE_MANAGER_DNF5),
    (CONTAINER_FEDORA_39, PACKAGE_MANAGER_DNF),
    (CONTAINER_FEDORA_39, PACKAGE_MANAGER_YUM),

    # CentOS Stream
    (CONTAINER_CENTOS_STREAM_9, PACKAGE_MANAGER_DNF),
    (CONTAINER_CENTOS_STREAM_9, PACKAGE_MANAGER_YUM),

    # CentOS
    (CONTAINER_CENTOS_7, PACKAGE_MANAGER_YUM),

    # UBI
    (CONTAINER_UBI_8, PACKAGE_MANAGER_DNF),
    (CONTAINER_UBI_8, PACKAGE_MANAGER_YUM),

    # Ubuntu
    (CONTAINER_UBUNTU_2204, PACKAGE_MANAGER_APT),

    # Debian
    (CONTAINER_DEBIAN_127, PACKAGE_MANAGER_APT),

    # Fedora CoreOS
    (CONTAINER_FEDORA_COREOS, PACKAGE_MANAGER_DNF5),
    (CONTAINER_FEDORA_COREOS_OSTREE, PACKAGE_MANAGER_RPMOSTREE),

    # Alpine
    (CONTAINER_ALPINE, PACKAGE_MANAGER_APK),
    ]

CONTAINER_MATRIX_IDS = [
    f'{container.url} / {package_manager_class.__name__.lower()}'
    for container, package_manager_class in CONTAINER_BASE_MATRIX
    ]


CONTAINER_DISCOVERY_MATRIX: dict[str, tuple[Container, PackageManagerClass]] = {}

for container, package_manager_class in CONTAINER_BASE_MATRIX:
    if container.url in CONTAINER_DISCOVERY_MATRIX:
        continue

    CONTAINER_DISCOVERY_MATRIX[container.url] = (container, package_manager_class)


@pytest.fixture(name='guest')
def fixture_guest(container: ContainerData, root_logger: tmt.log.Logger) -> GuestContainer:
    guest_data = PodmanGuestData(
        image=container.image_url_or_id,
        container=container.container_id
        )

    guest = GuestContainer(
        logger=root_logger,
        data=guest_data,
        name='dummy-container')

    guest.start()

    return guest


@pytest.fixture(name='guest_per_test')
def fixture_guest_per_test(
        container_per_test: ContainerData,
        root_logger: tmt.log.Logger) -> GuestContainer:
    guest_data = PodmanGuestData(
        image=container_per_test.image_url_or_id,
        container=container_per_test.container_id
        )

    guest = GuestContainer(
        logger=root_logger,
        data=guest_data,
        name='dummy-container')

    guest.start()

    return guest


def is_dnf5_preinstalled(container: ContainerData) -> bool:
    return container.image_url_or_id in (
        CONTAINER_FEDORA_RAWHIDE.url,
        CONTAINER_FEDORA_41.url,
        CONTAINER_FEDORA_COREOS.url,
        CONTAINER_FEDORA_COREOS_OSTREE.url)


def create_package_manager(
        container: ContainerData,
        guest: GuestContainer,
        package_manager_class: PackageManagerClass,
        logger: tmt.log.Logger) -> PackageManager:
    guest_data = tmt.steps.provision.podman.PodmanGuestData(
        image=container.image_url_or_id,
        container=container.container_id
        )

    guest = tmt.steps.provision.podman.GuestContainer(
        logger=logger,
        data=guest_data,
        name='dummy-container')
    guest.start()
    guest.show()

    if package_manager_class is tmt.package_managers.dnf.Dnf5:
        # Note that our custom images contain `dnf5` already
        if is_dnf5_preinstalled(container):
            pass

        else:
            guest.execute(ShellScript('dnf install --nogpgcheck -y dnf5'))

    return package_manager_class(guest=guest, logger=logger)


def _parametrize_test_discovery() -> Iterator[tuple[ContainerData, PackageManagerClass]]:
    yield from CONTAINER_DISCOVERY_MATRIX.values()


@pytest.mark.containers
@pytest.mark.parametrize(('container',
                          'expected_package_manager'),
                         list(_parametrize_test_discovery()),
                         indirect=["container"],
                         ids=[
                             container.url
                             for container, _
                             in CONTAINER_DISCOVERY_MATRIX.values()
    ])
def test_discovery(
        container: ContainerData,
        guest: GuestContainer,
        expected_package_manager: PackageManagerClass,
        root_logger: tmt.log.Logger,
        caplog: _pytest.logging.LogCaptureFixture) -> None:

    guest.show()

    def _test_discovery(expected: str, expected_discovery: str) -> None:
        caplog.clear()

        guest.facts.sync(guest)

        assert guest.facts.package_manager == expected

        assert_log(caplog, message=MATCH(
            rf"^Discovered package managers: {expected_discovery}$"))

    # Images in which `dnf5`` would be the best possible choice, do not
    # come with `dnf5`` pre-installed. Therefore run the discovery first,
    # but expect to find *dnf* instead of `dnf5`. Then install `dnf5`,
    # re-run the discovery and expect the original outcome.
    if expected_package_manager is tmt.package_managers.dnf.Dnf5:
        guest.info(f'{container.image_url_or_id=}')
        guest.info(f'{has_legacy_dnf(container)=}')
        guest.info(f'{has_dnf5_preinstalled(container)=}')

        if has_legacy_dnf(container):
            _test_discovery(tmt.package_managers.dnf.Dnf.NAME, 'dnf and yum')

            if not has_dnf5_preinstalled(container):
                guest.execute(ShellScript('dnf install --nogpgcheck -y dnf5'))

        if has_legacy_dnf(container):
            _test_discovery(expected_package_manager.NAME, 'dnf5, dnf and yum')

        else:
            _test_discovery(expected_package_manager.NAME, 'dnf5')

    elif expected_package_manager is tmt.package_managers.dnf.Dnf:
        _test_discovery(expected_package_manager.NAME, 'dnf and yum')

    elif expected_package_manager is tmt.package_managers.rpm_ostree.RpmOstree:
        _test_discovery(expected_package_manager.NAME, 'rpm-ostree and dnf5')

    else:
        _test_discovery(expected_package_manager.NAME, expected_package_manager.NAME)


def _parametrize_test_install() -> \
        Iterator[tuple[
            Container,
            PackageManagerClass,
            Package,
            str,
            Optional[str]]]:

    for container, package_manager_class in CONTAINER_BASE_MATRIX:
        if package_manager_class is tmt.package_managers.dnf.Yum:
            if container.url == CONTAINER_FEDORA_RAWHIDE.url:
                yield container, \
                    package_manager_class, \
                    Package('tree'), \
                    r"rpm -q --whatprovides tree \|\| yum install -y  tree && rpm -q --whatprovides tree", \
                    'Installing:'  # noqa: E501

            elif 'ubi/8' in container.url:
                yield container, \
                    package_manager_class, \
                    Package('dconf'), \
                    r"rpm -q --whatprovides dconf \|\| yum install -y  dconf && rpm -q --whatprovides dconf", \
                    'Installed:\n  dconf'  # noqa: E501

            else:
                yield container, \
                    package_manager_class, \
                    Package('tree'), \
                    r"rpm -q --whatprovides tree \|\| yum install -y  tree && rpm -q --whatprovides tree", \
                    'Installed:\n  tree'  # noqa: E501

        elif package_manager_class is tmt.package_managers.dnf.Dnf:
            if container.url == CONTAINER_FEDORA_RAWHIDE.url:
                yield container, \
                    package_manager_class, \
                    Package('tree'), \
                    r"rpm -q --whatprovides tree \|\| dnf install -y  tree", \
                    'Installing:'

            elif 'ubi/8' in container.url:
                yield container, \
                    package_manager_class, \
                    Package('dconf'), \
                    r"rpm -q --whatprovides dconf \|\| dnf install -y  dconf", \
                    'Installed:\n  dconf'

            else:
                yield container, \
                    package_manager_class, \
                    Package('tree'), \
                    r"rpm -q --whatprovides tree \|\| dnf install -y  tree", \
                    'Installed:\n  tree'

        elif package_manager_class is tmt.package_managers.dnf.Dnf5:
            yield container, \
                package_manager_class, \
                Package('tree'), \
                r"rpm -q --whatprovides tree \|\| dnf5 install -y  tree", \
                'Installing:'

        elif package_manager_class is tmt.package_managers.apt.Apt:
            yield container, \
                package_manager_class, \
                Package('tree'), \
                r"export DEBIAN_FRONTEND=noninteractive; dpkg-query --show tree \|\| apt install -y  tree", \
                'Setting up tree'  # noqa: E501

        elif package_manager_class is tmt.package_managers.rpm_ostree.RpmOstree:
            yield container, \
                package_manager_class, \
                Package('tree'), \
                r"rpm -q --whatprovides tree \|\| rpm-ostree install --apply-live --idempotent --allow-inactive  tree", \
                'Installing: tree'  # noqa: E501

        elif package_manager_class is tmt.package_managers.apk.Apk:
            yield container, \
                package_manager_class, \
                Package('tree'), \
                r"apk info -e tree \|\| apk add tree", \
                'Installing tree'

        else:
            pytest.fail(f"Unhandled package manager class '{package_manager_class}'.")


@pytest.mark.containers
@pytest.mark.parametrize(('container_per_test',
                          'package_manager_class',
                          'package',
                          'expected_command',
                          'expected_output'),
                         list(_parametrize_test_install()),
                         indirect=["container_per_test"],
                         ids=CONTAINER_MATRIX_IDS)
def test_install(
        container_per_test: ContainerData,
        guest_per_test: GuestContainer,
        package_manager_class: PackageManagerClass,
        package: Package,
        expected_command: str,
        expected_output: Optional[str],
        root_logger: tmt.log.Logger,
        caplog: _pytest.logging.LogCaptureFixture) -> None:
    package_manager = create_package_manager(
        container_per_test,
        guest_per_test,
        package_manager_class,
        root_logger)

    output = package_manager.install(package)

    assert_log(caplog, message=MATCH(
        rf"Run command: podman exec .+? /bin/bash -c '{expected_command}'"))

    assert_output(expected_output, output.stdout, output.stderr)


def _parametrize_test_install_nonexistent() -> \
        Iterator[tuple[Container, PackageManagerClass, str, Optional[str]]]:

    for container, package_manager_class in CONTAINER_BASE_MATRIX:
        if package_manager_class is tmt.package_managers.dnf.Dnf5:
            yield container, \
                package_manager_class, \
                r"rpm -q --whatprovides tree-but-spelled-wrong \|\| dnf5 install -y  tree-but-spelled-wrong", \
                'No match for argument: tree-but-spelled-wrong'  # noqa: E501

        elif package_manager_class is tmt.package_managers.dnf.Dnf:
            if container.url == CONTAINER_FEDORA_RAWHIDE.url:
                yield container, \
                    package_manager_class, \
                    r"rpm -q --whatprovides tree-but-spelled-wrong \|\| dnf install -y  tree-but-spelled-wrong", \
                    'No match for argument: tree-but-spelled-wrong'  # noqa: E501

            else:
                yield container, \
                    package_manager_class, \
                    r"rpm -q --whatprovides tree-but-spelled-wrong \|\| dnf install -y  tree-but-spelled-wrong", \
                    'Error: Unable to find a match: tree-but-spelled-wrong'  # noqa: E501

        elif package_manager_class is tmt.package_managers.dnf.Yum:
            if container.url == CONTAINER_FEDORA_RAWHIDE.url:
                yield container, \
                    package_manager_class, \
                    r"rpm -q --whatprovides tree-but-spelled-wrong \|\| yum install -y  tree-but-spelled-wrong && rpm -q --whatprovides tree-but-spelled-wrong", \
                    'No match for argument: tree-but-spelled-wrong'  # noqa: E501

            elif 'fedora' in container.url:
                yield container, \
                    package_manager_class, \
                    r"rpm -q --whatprovides tree-but-spelled-wrong \|\| yum install -y  tree-but-spelled-wrong && rpm -q --whatprovides tree-but-spelled-wrong", \
                    'Error: Unable to find a match: tree-but-spelled-wrong'  # noqa: E501

            elif ('centos' in container.url and 'centos/7' not in container.url) \
                    or 'ubi/8' in container.url:
                yield container, \
                    package_manager_class, \
                    r"rpm -q --whatprovides tree-but-spelled-wrong \|\| yum install -y  tree-but-spelled-wrong && rpm -q --whatprovides tree-but-spelled-wrong", \
                    'No match for argument: tree-but-spelled-wrong'  # noqa: E501

            else:
                yield container, \
                    package_manager_class, \
                    r"rpm -q --whatprovides tree-but-spelled-wrong \|\| yum install -y  tree-but-spelled-wrong && rpm -q --whatprovides tree-but-spelled-wrong", \
                    'No package tree-but-spelled-wrong available.'  # noqa: E501

        elif package_manager_class is tmt.package_managers.apt.Apt:
            yield container, \
                package_manager_class, \
                r"export DEBIAN_FRONTEND=noninteractive; dpkg-query --show tree-but-spelled-wrong \|\| apt install -y  tree-but-spelled-wrong", \
                'E: Unable to locate package tree-but-spelled-wrong'  # noqa: E501

        elif package_manager_class is tmt.package_managers.rpm_ostree.RpmOstree:
            yield container, \
                package_manager_class, \
                r"rpm -q --whatprovides tree-but-spelled-wrong \|\| rpm-ostree install --apply-live --idempotent --allow-inactive  tree-but-spelled-wrong", \
                'no package provides tree-but-spelled-wrong'  # noqa: E501

        elif package_manager_class is tmt.package_managers.apk.Apk:
            yield container, \
                package_manager_class, \
                r"apk info -e tree-but-spelled-wrong \|\| apk add tree-but-spelled-wrong", \
                'ERROR: unable to select packages:\n  tree-but-spelled-wrong (no such package):\n    required by: world[tree-but-spelled-wrong]'  # noqa: E501

        else:
            pytest.fail(f"Unhandled package manager class '{package_manager_class}'.")


@pytest.mark.containers
@pytest.mark.parametrize(('container',
                          'package_manager_class',
                          'expected_command',
                          'expected_output'),
                         list(_parametrize_test_install_nonexistent()),
                         indirect=["container"],
                         ids=CONTAINER_MATRIX_IDS)
def test_install_nonexistent(
        container: ContainerData,
        guest: GuestContainer,
        package_manager_class: PackageManagerClass,
        expected_command: str,
        expected_output: Optional[str],
        root_logger: tmt.log.Logger,
        caplog: _pytest.logging.LogCaptureFixture) -> None:
    package_manager = create_package_manager(container, guest, package_manager_class, root_logger)

    with pytest.raises(tmt.utils.RunError) as excinfo:
        package_manager.install(Package('tree-but-spelled-wrong'))

    assert_log(caplog, message=MATCH(
        rf"Run command: podman exec .+? /bin/bash -c '{expected_command}'"))

    assert excinfo.type is tmt.utils.RunError
    assert excinfo.value.returncode != 0

    assert_output(expected_output, excinfo.value.stdout, excinfo.value.stderr)


def _parametrize_test_install_nonexistent_skip() -> \
        Iterator[tuple[Container, PackageManagerClass, str, Optional[str]]]:

    for container, package_manager_class in CONTAINER_BASE_MATRIX:
        if package_manager_class is tmt.package_managers.dnf.Dnf5:
            yield container, \
                package_manager_class, \
                r"rpm -q --whatprovides tree-but-spelled-wrong \|\| dnf5 install -y --skip-unavailable tree-but-spelled-wrong", \
                None  # noqa: E501

        elif package_manager_class is tmt.package_managers.dnf.Dnf:
            yield container, \
                package_manager_class, \
                r"rpm -q --whatprovides tree-but-spelled-wrong \|\| dnf install -y --skip-broken tree-but-spelled-wrong", \
                'No match for argument: tree-but-spelled-wrong'  # noqa: E501

        elif package_manager_class is tmt.package_managers.dnf.Yum:
            if container.url == CONTAINER_FEDORA_RAWHIDE.url:  # noqa: SIM114
                yield container, \
                    package_manager_class, \
                    r"rpm -q --whatprovides tree-but-spelled-wrong \|\| yum install -y --skip-broken tree-but-spelled-wrong \|\| /bin/true", \
                    'No match for argument: tree-but-spelled-wrong'  # noqa: E501

            elif 'fedora' in container.url:  # noqa: SIM114
                yield container, \
                    package_manager_class, \
                    r"rpm -q --whatprovides tree-but-spelled-wrong \|\| yum install -y --skip-broken tree-but-spelled-wrong \|\| /bin/true", \
                    'No match for argument: tree-but-spelled-wrong'  # noqa: E501

            elif ('centos' in container.url and 'centos/7' not in container.url) \
                    or 'ubi/8' in container.url:
                yield container, \
                    package_manager_class, \
                    r"rpm -q --whatprovides tree-but-spelled-wrong \|\| yum install -y --skip-broken tree-but-spelled-wrong \|\| /bin/true", \
                    'No match for argument: tree-but-spelled-wrong'  # noqa: E501

            else:
                yield container, \
                    package_manager_class, \
                    r"rpm -q --whatprovides tree-but-spelled-wrong \|\| yum install -y --skip-broken tree-but-spelled-wrong \|\| /bin/true", \
                    'No package tree-but-spelled-wrong available.'  # noqa: E501

        elif package_manager_class is tmt.package_managers.apt.Apt:
            yield container, \
                package_manager_class, \
                r"export DEBIAN_FRONTEND=noninteractive; dpkg-query --show tree-but-spelled-wrong \|\| apt install -y --ignore-missing tree-but-spelled-wrong \|\| /bin/true", \
                'E: Unable to locate package tree-but-spelled-wrong'  # noqa: E501

        elif package_manager_class is tmt.package_managers.rpm_ostree.RpmOstree:
            yield container, \
                package_manager_class, \
                r"rpm -q --whatprovides tree-but-spelled-wrong \|\| rpm-ostree install --apply-live --idempotent --allow-inactive  tree-but-spelled-wrong \|\| /bin/true", \
                'no package provides tree-but-spelled-wrong'  # noqa: E501

        elif package_manager_class is tmt.package_managers.apk.Apk:
            yield container, \
                package_manager_class, \
                r"apk info -e tree-but-spelled-wrong \|\| apk add tree-but-spelled-wrong \|\| /bin/true", \
                'ERROR: unable to select packages:\n  tree-but-spelled-wrong (no such package):\n    required by: world[tree-but-spelled-wrong]'  # noqa: E501

        else:
            pytest.fail(f"Unhandled package manager class '{package_manager_class}'.")


@pytest.mark.containers
@pytest.mark.parametrize(('container',
                          'package_manager_class',
                          'expected_command',
                          'expected_output'),
                         list(_parametrize_test_install_nonexistent_skip()),
                         indirect=["container"],
                         ids=CONTAINER_MATRIX_IDS)
def test_install_nonexistent_skip(
        container: ContainerData,
        guest: GuestContainer,
        package_manager_class: PackageManagerClass,
        expected_command: str,
        expected_output: Optional[str],
        root_logger: tmt.log.Logger,
        caplog: _pytest.logging.LogCaptureFixture) -> None:
    package_manager = create_package_manager(container, guest, package_manager_class, root_logger)

    output = package_manager.install(
        Package('tree-but-spelled-wrong'),
        options=Options(skip_missing=True)
        )

    assert_log(caplog, message=MATCH(
        rf"Run command: podman exec .+? /bin/bash -c '{expected_command}'"))

    assert_output(expected_output, output.stdout, output.stderr)


def _parametrize_test_install_dont_check_first() -> \
        Iterator[tuple[
            Container,
            PackageManagerClass,
            Package,
            str,
            Optional[str]]]:

    for container, package_manager_class in CONTAINER_BASE_MATRIX:
        if package_manager_class is tmt.package_managers.dnf.Dnf5:
            yield container, \
                package_manager_class, \
                Package('tree'), \
                r"dnf5 install -y  tree", \
                None

        elif package_manager_class is tmt.package_managers.dnf.Dnf:
            if 'ubi/8' in container.url:
                yield container, \
                    package_manager_class, \
                    Package('dconf'), \
                    r"dnf install -y  dconf", \
                    'Installed:\n  dconf'
            else:
                yield container, \
                    package_manager_class, \
                    Package('tree'), \
                    r"dnf install -y  tree", \
                    'Installed:\n  tree'

        elif package_manager_class is tmt.package_managers.dnf.Yum:
            if 'ubi/8' in container.url:
                yield container, \
                    package_manager_class, \
                    Package('dconf'), \
                    r"yum install -y  dconf && rpm -q --whatprovides dconf", \
                    'Installed:\n  dconf'

            else:
                yield container, \
                    package_manager_class, \
                    Package('tree'), \
                    r"yum install -y  tree && rpm -q --whatprovides tree", \
                    'Installed:\n  tree'

        elif package_manager_class is tmt.package_managers.apt.Apt:
            yield container, \
                package_manager_class, \
                Package('tree'), \
                r"export DEBIAN_FRONTEND=noninteractive; apt install -y  tree", \
                'Setting up tree'

        elif package_manager_class is tmt.package_managers.rpm_ostree.RpmOstree:
            yield container, \
                package_manager_class, \
                Package('tree'), \
                r"rpm-ostree install --apply-live --idempotent --allow-inactive  tree", \
                'Installing: tree'

        elif package_manager_class is tmt.package_managers.apk.Apk:
            yield container, \
                package_manager_class, \
                Package('tree'), \
                r"apk add tree", \
                'Installing tree'

        else:
            pytest.fail(f"Unhandled package manager class '{package_manager_class}'.")


@pytest.mark.containers
@pytest.mark.parametrize(('container_per_test',
                          'package_manager_class',
                          'package',
                          'expected_command',
                          'expected_output'),
                         list(_parametrize_test_install_dont_check_first()),
                         indirect=["container_per_test"],
                         ids=CONTAINER_MATRIX_IDS)
def test_install_dont_check_first(
        container_per_test: ContainerData,
        guest_per_test: GuestContainer,
        package_manager_class: PackageManagerClass,
        package: Package,
        expected_command: str,
        expected_output: Optional[str],
        root_logger: tmt.log.Logger,
        caplog: _pytest.logging.LogCaptureFixture) -> None:
    package_manager = create_package_manager(
        container_per_test,
        guest_per_test,
        package_manager_class,
        root_logger)

    output = package_manager.install(
        package,
        options=Options(check_first=False)
        )

    assert_log(caplog, message=MATCH(
        rf"Run command: podman exec .+? /bin/bash -c '{expected_command}'"))

    assert_output(expected_output, output.stdout, output.stderr)


def _parametrize_test_reinstall() -> Iterator[tuple[
        Container, PackageManagerClass, Optional[str], Optional[str]]]:

    for container, package_manager_class in CONTAINER_BASE_MATRIX:
        if package_manager_class is tmt.package_managers.dnf.Yum:
            if 'centos/7' in container.url:
                yield container, \
                    package_manager_class, \
                    Package('tar'), \
                    True, \
                    r"rpm -q --whatprovides tar && yum reinstall -y  tar && rpm -q --whatprovides tar", \
                    'Reinstalling:\n tar'  # noqa: E501

            else:
                yield container, \
                    package_manager_class, \
                    Package('tar'), \
                    True, \
                    r"rpm -q --whatprovides tar && yum reinstall -y  tar && rpm -q --whatprovides tar", \
                    'Reinstalled:\n  tar'  # noqa: E501

        elif package_manager_class is tmt.package_managers.dnf.Dnf:
            yield container, \
                package_manager_class, \
                Package('tar'), \
                True, \
                r"rpm -q --whatprovides tar && dnf reinstall -y  tar", \
                'Reinstalled:\n  tar'

        elif package_manager_class is tmt.package_managers.dnf.Dnf5:
            yield container, \
                package_manager_class, \
                Package('tar'), \
                True, \
                r"rpm -q --whatprovides tar && dnf5 reinstall -y  tar", \
                'Reinstalling tar'

        elif package_manager_class is tmt.package_managers.apt.Apt:
            yield container, \
                package_manager_class, \
                Package('tar'), \
                True, \
                r"export DEBIAN_FRONTEND=noninteractive; dpkg-query --show tar && apt reinstall -y  tar", \
                'Setting up tar'  # noqa: E501

        elif package_manager_class is tmt.package_managers.rpm_ostree.RpmOstree:
            yield container, \
                package_manager_class, \
                Package('tar'), \
                False, \
                None, \
                None

        elif package_manager_class is tmt.package_managers.apk.Apk:
            yield container, \
                package_manager_class, \
                Package('bash'), \
                True, \
                r"apk info -e bash && apk fix bash", \
                'Reinstalling bash'

        else:
            pytest.fail(f"Unhandled package manager class '{package_manager_class}'.")


@pytest.mark.containers
@pytest.mark.parametrize(('container_per_test',
                          'package_manager_class',
                          'package',
                          'supported',
                          'expected_command',
                          'expected_output'),
                         list(_parametrize_test_reinstall()),
                         indirect=["container_per_test"],
                         ids=CONTAINER_MATRIX_IDS)
def test_reinstall(
        container_per_test: ContainerData,
        guest_per_test: GuestContainer,
        package_manager_class: PackageManagerClass,
        package: Package,
        supported: bool,
        expected_command: Optional[str],
        expected_output: Optional[str],
        root_logger: tmt.log.Logger,
        caplog: _pytest.logging.LogCaptureFixture) -> None:
    package_manager = create_package_manager(
        container_per_test,
        guest_per_test,
        package_manager_class,
        root_logger)

    if supported:
        assert expected_command is not None

        output = package_manager.reinstall(package)

        assert_log(caplog, message=MATCH(
            rf"Run command: podman exec .+? /bin/bash -c '{expected_command}'"))

    else:
        with pytest.raises(tmt.utils.GeneralError) as excinfo:
            package_manager.reinstall(package)

        assert excinfo.value.message \
            == "rpm-ostree does not support reinstall operation."

    if expected_output:
        assert_output(expected_output, output.stdout, output.stderr)


def _generate_test_reinstall_nonexistent_matrix() -> Iterator[tuple[
        Container, PackageManagerClass, Optional[str], Optional[str], Optional[str]]]:

    for container, package_manager_class in CONTAINER_BASE_MATRIX:
        if package_manager_class is tmt.package_managers.dnf.Dnf5:
            yield container, \
                package_manager_class, \
                True, \
                r"rpm -q --whatprovides tree-but-spelled-wrong && dnf5 reinstall -y  tree-but-spelled-wrong", \
                'no package provides tree-but-spelled-wrong'  # noqa: E501

        elif package_manager_class is tmt.package_managers.dnf.Dnf:
            yield container, \
                package_manager_class, \
                True, \
                r"rpm -q --whatprovides tree-but-spelled-wrong && dnf reinstall -y  tree-but-spelled-wrong", \
                'no package provides tree-but-spelled-wrong'  # noqa: E501

        elif package_manager_class is tmt.package_managers.dnf.Yum:
            if 'fedora' in container.url:  # noqa: SIM114
                yield container, \
                    package_manager_class, \
                    True, \
                    r"rpm -q --whatprovides tree-but-spelled-wrong && yum reinstall -y  tree-but-spelled-wrong && rpm -q --whatprovides tree-but-spelled-wrong", \
                    'no package provides tree-but-spelled-wrong'  # noqa: E501

            elif 'centos' in container.url and 'centos/7' not in container.url:
                yield container, \
                    package_manager_class, \
                    True, \
                    r"rpm -q --whatprovides tree-but-spelled-wrong && yum reinstall -y  tree-but-spelled-wrong && rpm -q --whatprovides tree-but-spelled-wrong", \
                    'no package provides tree-but-spelled-wrong'  # noqa: E501

            else:
                yield container, \
                    package_manager_class, \
                    True, \
                    r"rpm -q --whatprovides tree-but-spelled-wrong && yum reinstall -y  tree-but-spelled-wrong && rpm -q --whatprovides tree-but-spelled-wrong", \
                    'no package provides tree-but-spelled-wrong'  # noqa: E501

        elif package_manager_class is tmt.package_managers.apt.Apt:
            yield container, \
                package_manager_class, \
                True, \
                r"export DEBIAN_FRONTEND=noninteractive; dpkg-query --show tree-but-spelled-wrong && apt reinstall -y  tree-but-spelled-wrong", \
                'dpkg-query: no packages found matching tree-but-spelled-wrong'  # noqa: E501

        elif package_manager_class is tmt.package_managers.rpm_ostree.RpmOstree:
            yield container, \
                package_manager_class, \
                False, \
                None, \
                None

        elif package_manager_class is tmt.package_managers.apk.Apk:
            yield container, \
                package_manager_class, \
                True, \
                r"apk info -e tree-but-spelled-wrong && apk fix tree-but-spelled-wrong", \
                None

        else:
            pytest.fail(f"Unhandled package manager class '{package_manager_class}'.")


@pytest.mark.containers
@pytest.mark.parametrize(('container',
                          'package_manager_class',
                          'supported',
                          'expected_command',
                          'expected_output'),
                         list(_generate_test_reinstall_nonexistent_matrix()),
                         indirect=["container"],
                         ids=CONTAINER_MATRIX_IDS)
def test_reinstall_nonexistent(
        container: ContainerData,
        guest: GuestContainer,
        package_manager_class: PackageManagerClass,
        supported: bool,
        expected_command: str,
        expected_output: Optional[str],
        root_logger: tmt.log.Logger,
        caplog: _pytest.logging.LogCaptureFixture) -> None:
    package_manager = create_package_manager(container, guest, package_manager_class, root_logger)

    if supported:
        assert expected_command is not None

        with pytest.raises(tmt.utils.RunError) as excinfo:
            package_manager.reinstall(Package('tree-but-spelled-wrong'))

        assert_log(caplog, message=MATCH(
            rf"Run command: podman exec .+? /bin/bash -c '{expected_command}'"))

        assert excinfo.type is tmt.utils.RunError
        assert excinfo.value.returncode != 0

    else:
        with pytest.raises(tmt.utils.GeneralError) as excinfo:
            package_manager.reinstall(Package('tree-but-spelled-wrong'))

        assert excinfo.value.message \
            == "rpm-ostree does not support reinstall operation."

    if expected_output:
        assert_output(expected_output, excinfo.value.stdout, excinfo.value.stderr)


def _generate_test_check_presence() -> Iterator[
        tuple[Container, PackageManagerClass, Installable, str, Optional[str]]]:

    for container, package_manager_class in CONTAINER_BASE_MATRIX:
        if package_manager_class is tmt.package_managers.dnf.Dnf5:
            yield container, \
                package_manager_class, \
                Package('coreutils'), \
                True, \
                r"rpm -q --whatprovides coreutils", \
                r'\s+out:\s+coreutils-'

            yield container, \
                package_manager_class, \
                Package('tree-but-spelled-wrong'), \
                False, \
                r"rpm -q --whatprovides tree-but-spelled-wrong", \
                r'\s+out:\s+no package provides tree-but-spelled-wrong'

            yield container, \
                package_manager_class, \
                FileSystemPath('/usr/bin/arch'), \
                True, \
                r"rpm -q --whatprovides /usr/bin/arch", \
                r'\s+out:\s+coreutils-'

        elif package_manager_class is tmt.package_managers.dnf.Dnf:
            if 'ubi/8' in container.url:
                yield container, \
                    package_manager_class, \
                    Package('util-linux'), \
                    True, \
                    r"rpm -q --whatprovides util-linux", \
                    r'\s+out:\s+util-linux-'

                yield container, \
                    package_manager_class, \
                    Package('tree-but-spelled-wrong'), \
                    False, \
                    r"rpm -q --whatprovides tree-but-spelled-wrong", \
                    r'\s+out:\s+no package provides tree-but-spelled-wrong'

                yield container, \
                    package_manager_class, \
                    FileSystemPath('/usr/bin/flock'), \
                    True, \
                    r"rpm -q --whatprovides /usr/bin/flock", \
                    r'\s+out:\s+util-linux-'

            elif 'centos/stream9' in container.url or 'fedora/40' in container.url:
                yield container, \
                    package_manager_class, \
                    Package('coreutils'), \
                    True, \
                    r"rpm -q --whatprovides coreutils", \
                    r'\s+out:\s+coreutils-'

                yield container, \
                    package_manager_class, \
                    Package('tree-but-spelled-wrong'), \
                    False, \
                    r"rpm -q --whatprovides tree-but-spelled-wrong", \
                    r'\s+out:\s+no package provides tree-but-spelled-wrong'

                yield container, \
                    package_manager_class, \
                    FileSystemPath('/usr/bin/arch'), \
                    True, \
                    r"rpm -q --whatprovides /usr/bin/arch", \
                    r'\s+out:\s+coreutils-'

            else:
                yield container, \
                    package_manager_class, \
                    Package('util-linux-core'), \
                    True, \
                    r"rpm -q --whatprovides util-linux-core", \
                    r'\s+out:\s+util-linux-core-'

                yield container, \
                    package_manager_class, \
                    Package('tree-but-spelled-wrong'), \
                    False, \
                    r"rpm -q --whatprovides tree-but-spelled-wrong", \
                    r'\s+out:\s+no package provides tree-but-spelled-wrong'

                yield container, \
                    package_manager_class, \
                    FileSystemPath('/usr/bin/flock'), \
                    True, \
                    r"rpm -q --whatprovides /usr/bin/flock", \
                    r'\s+out:\s+util-linux-core-'

        elif package_manager_class is tmt.package_managers.dnf.Yum:
            if 'centos/7' in container.url or 'ubi/8' in container.url:
                yield container, \
                    package_manager_class, \
                    Package('util-linux'), \
                    True, \
                    r"rpm -q --whatprovides util-linux", \
                    r'\s+out:\s+util-linux-'

                yield container, \
                    package_manager_class, \
                    Package('tree-but-spelled-wrong'), \
                    False, \
                    r"rpm -q --whatprovides tree-but-spelled-wrong", \
                    r'\s+out:\s+no package provides tree-but-spelled-wrong'

                yield container, \
                    package_manager_class, \
                    FileSystemPath('/usr/bin/flock'), \
                    True, \
                    r"rpm -q --whatprovides /usr/bin/flock", \
                    r'\s+out:\s+util-linux-'

            elif 'centos/stream9' in container.url or 'fedora/40' in container.url:
                yield container, \
                    package_manager_class, \
                    Package('coreutils'), \
                    True, \
                    r"rpm -q --whatprovides coreutils", \
                    r'\s+out:\s+coreutils-'

                yield container, \
                    package_manager_class, \
                    Package('tree-but-spelled-wrong'), \
                    False, \
                    r"rpm -q --whatprovides tree-but-spelled-wrong", \
                    r'\s+out:\s+no package provides tree-but-spelled-wrong'

                yield container, \
                    package_manager_class, \
                    FileSystemPath('/usr/bin/arch'), \
                    True, \
                    r"rpm -q --whatprovides /usr/bin/arch", \
                    r'\s+out:\s+coreutils-'

            else:
                yield container, \
                    package_manager_class, \
                    Package('util-linux-core'), \
                    True, \
                    r"rpm -q --whatprovides util-linux-core", \
                    r'\s+out:\s+util-linux-core-'

                yield container, \
                    package_manager_class, \
                    Package('tree-but-spelled-wrong'), \
                    False, \
                    r"rpm -q --whatprovides tree-but-spelled-wrong", \
                    r'\s+out:\s+no package provides tree-but-spelled-wrong'

                yield container, \
                    package_manager_class, \
                    FileSystemPath('/usr/bin/flock'), \
                    True, \
                    r"rpm -q --whatprovides /usr/bin/flock", \
                    r'\s+out:\s+util-linux-core-'

        elif package_manager_class is tmt.package_managers.apt.Apt:
            yield container, \
                package_manager_class, \
                Package('util-linux'), \
                True, \
                r"dpkg-query --show util-linux", \
                r'\s+out:\s+util-linux'

            yield container, \
                package_manager_class, \
                Package('tree-but-spelled-wrong'), \
                False, \
                r"dpkg-query --show tree-but-spelled-wrong", \
                r'\s+err:\s+dpkg-query: no packages found matching tree-but-spelled-wrong'

            yield container, \
                package_manager_class, \
                FileSystemPath('/usr/bin/flock'), \
                True, \
                r"dpkg-query --show util-linux", \
                r'\s+out:\s+util-linux'

        elif package_manager_class is tmt.package_managers.rpm_ostree.RpmOstree:
            yield container, \
                package_manager_class, \
                Package('util-linux'), \
                True, \
                r"rpm -q --whatprovides util-linux", \
                r'\s+out:\s+util-linux'

            yield container, \
                package_manager_class, \
                Package('tree-but-spelled-wrong'), \
                False, \
                r"rpm -q --whatprovides tree-but-spelled-wrong", \
                r'\s+out:\s+no package provides tree-but-spelled-wrong'

            yield container, \
                package_manager_class, \
                FileSystemPath('/usr/bin/flock'), \
                True, \
                r"rpm -qf /usr/bin/flock", \
                r'\s+out:\s+util-linux-core'

        elif package_manager_class is tmt.package_managers.apk.Apk:
            yield container, \
                package_manager_class, \
                Package('busybox'), \
                True, \
                r"apk info -e busybox", \
                r'\s+out:\s+busybox'

            yield container, \
                package_manager_class, \
                Package('tree-but-spelled-wrong'), \
                False, \
                r"apk info -e tree-but-spelled-wrong", \
                None

            yield container, \
                package_manager_class, \
                FileSystemPath('/usr/bin/arch'), \
                True, \
                r"apk info -e busybox", \
                r'\s+out:\s+busybox'

        else:
            pytest.fail(f"Unhandled package manager class '{package_manager_class}'.")


def _generate_test_check_presence_ids(value) -> str:
    if isinstance(value, Container):
        return value.url

    if isclass(value) and issubclass(value, tmt.package_managers.PackageManager):
        return value.__name__.lower()

    if isinstance(value, (Package, FileSystemPath)):
        return str(value)

    return ''


@pytest.mark.containers
@pytest.mark.parametrize(('container',
                          'package_manager_class',
                          'installable',
                          'expected_result',
                          'expected_command',
                          'expected_output'),
                         list(_generate_test_check_presence()),
                         indirect=["container"],
                         ids=_generate_test_check_presence_ids)
def test_check_presence(
        container: ContainerData,
        guest: GuestContainer,
        package_manager_class: PackageManagerClass,
        installable: Installable,
        expected_result: bool,
        expected_command: str,
        expected_output: Optional[str],
        root_logger: tmt.log.Logger,
        caplog: _pytest.logging.LogCaptureFixture) -> None:
    package_manager = create_package_manager(container, guest, package_manager_class, root_logger)

    assert package_manager.check_presence(installable) == {installable: expected_result}

    assert_log(caplog, message=MATCH(
        rf"Run command: podman exec .+? /bin/bash -c '{expected_command}'"))

    if expected_output:
        assert_log(caplog, remove_colors=True, message=MATCH(expected_output))


def _parametrize_test_install_filesystempath() -> Iterator[
        tuple[Container, PackageManagerClass, FileSystemPath, Optional[str]]]:

    for container, package_manager_class in CONTAINER_BASE_MATRIX:
        if package_manager_class is tmt.package_managers.dnf.Dnf5:
            yield container, \
                package_manager_class, \
                FileSystemPath('/usr/bin/dos2unix'), \
                r"rpm -q --whatprovides /usr/bin/dos2unix \|\| dnf5 install -y  /usr/bin/dos2unix", \
                '[1/1] dos2unix'  # noqa: E501

        elif package_manager_class is tmt.package_managers.dnf.Dnf:
            yield container, \
                package_manager_class, \
                FileSystemPath('/usr/bin/dos2unix'), \
                r"rpm -q --whatprovides /usr/bin/dos2unix \|\| dnf install -y  /usr/bin/dos2unix", \
                'Installed:\n  dos2unix-'  # noqa: E501

        elif package_manager_class is tmt.package_managers.dnf.Yum:
            if 'centos/7' in container.url:
                yield container, \
                    package_manager_class, \
                    FileSystemPath('/usr/bin/dos2unix'), \
                    r"rpm -q --whatprovides /usr/bin/dos2unix \|\| yum install -y  /usr/bin/dos2unix && rpm -q --whatprovides /usr/bin/dos2unix", \
                    'Installed:\n  dos2unix.'  # noqa: E501

            else:
                yield container, \
                    package_manager_class, \
                    FileSystemPath('/usr/bin/dos2unix'), \
                    r"rpm -q --whatprovides /usr/bin/dos2unix \|\| yum install -y  /usr/bin/dos2unix && rpm -q --whatprovides /usr/bin/dos2unix", \
                    'Installed:\n  dos2unix-'  # noqa: E501

        elif package_manager_class is tmt.package_managers.apt.Apt:
            yield container, \
                package_manager_class, \
                FileSystemPath('/usr/bin/dos2unix'), \
                r"export DEBIAN_FRONTEND=noninteractive; dpkg-query --show dos2unix \|\| apt install -y  dos2unix", \
                "Setting up dos2unix"  # noqa: E501

        elif package_manager_class is tmt.package_managers.rpm_ostree.RpmOstree:
            yield container, \
                package_manager_class, \
                FileSystemPath('/usr/bin/dos2unix'), \
                r"rpm -qf /usr/bin/dos2unix \|\| rpm-ostree install --apply-live --idempotent --allow-inactive  /usr/bin/dos2unix", \
                "Installing 1 packages:\n  dos2unix-"  # noqa: E501

        elif package_manager_class is tmt.package_managers.apk.Apk:
            yield container, \
                package_manager_class, \
                FileSystemPath('/usr/bin/dos2unix'), \
                r"apk info -e dos2unix \|\| apk add dos2unix", \
                'Installing dos2unix'

        else:
            pytest.fail(f"Unhandled package manager class '{package_manager_class}'.")


@pytest.mark.containers
@pytest.mark.parametrize(('container_per_test',
                          'package_manager_class',
                          'installable',
                          'expected_command',
                          'expected_output'),
                         list(_parametrize_test_install_filesystempath()),
                         indirect=["container_per_test"],
                         ids=CONTAINER_MATRIX_IDS)
def test_install_filesystempath(
        container_per_test: ContainerData,
        guest_per_test: GuestContainer,
        package_manager_class: PackageManagerClass,
        installable: FileSystemPath,
        expected_command: str,
        expected_output: Optional[str],
        root_logger: tmt.log.Logger,
        caplog: _pytest.logging.LogCaptureFixture) -> None:
    package_manager = create_package_manager(
        container_per_test,
        guest_per_test,
        package_manager_class,
        root_logger)

    output = package_manager.install(installable)

    assert_log(caplog, message=MATCH(
        rf"Run command: podman exec .+? /bin/bash -c '{expected_command}'"))

    assert_output(expected_output, output.stdout, output.stderr)


def _parametrize_test_install_multiple() -> \
        Iterator[tuple[
            Container,
            PackageManagerClass,
            tuple[Package, Package],
            str,
            Optional[str]]]:

    for container, package_manager_class in CONTAINER_BASE_MATRIX:
        if package_manager_class is tmt.package_managers.dnf.Yum:
            if 'ubi/8' in container.url:
                yield container, \
                    package_manager_class, \
                    (Package('dconf'), Package('libpng')), \
                    r"rpm -q --whatprovides dconf libpng \|\| yum install -y  dconf libpng && rpm -q --whatprovides dconf libpng", \
                    'Complete!'  # noqa: E501

            else:
                yield container, \
                    package_manager_class, \
                    (Package('tree'), Package('diffutils')), \
                    r"rpm -q --whatprovides tree diffutils \|\| yum install -y  tree diffutils && rpm -q --whatprovides tree diffutils", \
                    'Complete!'  # noqa: E501

        elif package_manager_class is tmt.package_managers.dnf.Dnf:
            if 'ubi/8' in container.url:
                yield container, \
                    package_manager_class, \
                    (Package('dconf'), Package('libpng')), \
                    r"rpm -q --whatprovides dconf libpng \|\| dnf install -y  dconf libpng", \
                    'Complete!'

            else:
                yield container, \
                    package_manager_class, \
                    (Package('tree'), Package('diffutils')), \
                    r"rpm -q --whatprovides tree diffutils \|\| dnf install -y  tree diffutils", \
                    'Complete!'

        elif package_manager_class is tmt.package_managers.dnf.Dnf5:
            yield container, \
                package_manager_class, \
                (Package('tree'), Package('diffutils')), \
                r"rpm -q --whatprovides tree diffutils \|\| dnf5 install -y  tree diffutils", \
                None

        elif package_manager_class is tmt.package_managers.apt.Apt:
            yield container, \
                package_manager_class, \
                (Package('tree'), Package('diffutils')), \
                r"export DEBIAN_FRONTEND=noninteractive; dpkg-query --show tree diffutils \|\| apt install -y  tree diffutils", \
                'Setting up tree'  # noqa: E501

        elif package_manager_class is tmt.package_managers.rpm_ostree.RpmOstree:
            yield container, \
                package_manager_class, \
                (Package('tree'), Package('diffutils')), \
                r"rpm -q --whatprovides tree diffutils \|\| rpm-ostree install --apply-live --idempotent --allow-inactive  tree diffutils", \
                'Installing: tree'  # noqa: E501

        elif package_manager_class is tmt.package_managers.apk.Apk:
            yield container, \
                package_manager_class, \
                (Package('tree'), Package('diffutils')), \
                r"apk info -e tree diffutils \|\| apk add tree diffutils", \
                'Installing tree'

        else:
            pytest.fail(f"Unhandled package manager class '{package_manager_class}'.")


@pytest.mark.containers
@pytest.mark.parametrize(('container_per_test',
                          'package_manager_class',
                          'packages',
                          'expected_command',
                          'expected_output'),
                         list(_parametrize_test_install_multiple()),
                         indirect=["container_per_test"],
                         ids=CONTAINER_MATRIX_IDS)
def test_install_multiple(
        container_per_test: ContainerData,
        guest_per_test: GuestContainer,
        package_manager_class: PackageManagerClass,
        packages: tuple[Package, Package],
        expected_command: str,
        expected_output: Optional[str],
        root_logger: tmt.log.Logger,
        caplog: _pytest.logging.LogCaptureFixture) -> None:
    package_manager = create_package_manager(
        container_per_test,
        guest_per_test,
        package_manager_class,
        root_logger)

    output = package_manager.install(*packages)

    assert_log(caplog, message=MATCH(
        rf"Run command: podman exec .+? /bin/bash -c '{expected_command}'"))

    assert_output(expected_output, output.stdout, output.stderr)


def _parametrize_test_install_downloaded() -> \
        Iterator[tuple[
            Container,
            PackageManagerClass,
            tuple[Package, Package],
            str,
            Optional[str]]]:

    for container, package_manager_class in CONTAINER_BASE_MATRIX:
        if package_manager_class is tmt.package_managers.dnf.Yum:
            if 'centos/7' in container.url:
                yield pytest.param(
                    container,
                    package_manager_class,
                    (Package('tree'), Package('diffutils')),
                    r"yum install -y --skip-broken /tmp/tree.rpm /tmp/diffutils.rpm \|\| /bin/true",  # noqa: E501
                    'Complete!',
                    marks=pytest.mark.skip(reason="CentOS 7 does not support 'download' command")
                    )

            elif 'ubi/8' in container.url:
                yield container, \
                    package_manager_class, \
                    (Package('dconf'), Package('libpng')), \
                    r"yum install -y --skip-broken /tmp/dconf.rpm /tmp/libpng.rpm \|\| /bin/true", \
                    'Complete!'  # noqa: E501

            else:
                yield container, \
                    package_manager_class, \
                    (Package('tree'), Package('diffutils')), \
                    r"yum install -y --skip-broken /tmp/tree.rpm /tmp/diffutils.rpm \|\| /bin/true", \
                    'Complete!'  # noqa: E501

        elif package_manager_class is tmt.package_managers.dnf.Dnf:
            if 'ubi/8' in container.url:
                yield container, \
                    package_manager_class, \
                    (Package('dconf'), Package('libpng')), \
                    r"dnf install -y  /tmp/dconf.rpm /tmp/libpng.rpm", \
                    'Complete!'

            else:
                yield container, \
                    package_manager_class, \
                    (Package('tree'), Package('diffutils')), \
                    r"dnf install -y  /tmp/tree.rpm /tmp/diffutils.rpm", \
                    'Complete!'

        elif package_manager_class is tmt.package_managers.dnf.Dnf5:
            yield container, \
                package_manager_class, \
                (Package('tree'), Package('diffutils')), \
                r"dnf5 install -y  /tmp/tree.rpm /tmp/diffutils.rpm", \
                None

        elif package_manager_class is tmt.package_managers.rpm_ostree.RpmOstree:
            yield container, \
                package_manager_class, \
                (Package('tree'), Package('diffutils')), \
                r"rpm-ostree install --apply-live --idempotent --allow-inactive  /tmp/tree.rpm /tmp/diffutils.rpm", \
                'Installing: tree'  # noqa: E501

        elif package_manager_class is tmt.package_managers.apt.Apt:
            yield pytest.param(
                container,
                package_manager_class,
                (Package('tree'), Package('diffutils')),
                r"export DEBIAN_FRONTEND=noninteractive; dpkg-query --show tree diffutils \|\| apt install -y  tree diffutils",  # noqa: E501
                'Setting up tree',
                marks=pytest.mark.skip(reason="not supported yet")
            )

        elif package_manager_class is tmt.package_managers.apk.Apk:
            yield pytest.param(
                container,
                package_manager_class,
                (Package('tree'), Package('diffutils')),
                r"apk info -e tree diffutils \|\| apk add tree diffutils",
                'Installing tree',
                marks=pytest.mark.skip(reason="not supported yet")
                )

        else:
            pytest.fail(f"Unhandled package manager class '{package_manager_class}'.")


@pytest.mark.containers
@pytest.mark.parametrize(('container_per_test',
                          'package_manager_class',
                          'packages',
                          'expected_command',
                          'expected_output'),
                         list(_parametrize_test_install_downloaded()),
                         indirect=["container_per_test"],
                         ids=CONTAINER_MATRIX_IDS)
def test_install_downloaded(
        container_per_test: ContainerData,
        guest_per_test: GuestContainer,
        package_manager_class: PackageManagerClass,
        packages: tuple[Package, Package],
        expected_command: str,
        expected_output: Optional[str],
        root_logger: tmt.log.Logger,
        caplog: _pytest.logging.LogCaptureFixture) -> None:
    package_manager = create_package_manager(
        container_per_test,
        guest_per_test,
        package_manager_class,
        root_logger)

    installables = tuple(PackagePath(f'/tmp/{package}.rpm') for package in packages)

    # TODO: move to a fixture
    guest_per_test.execute(ShellScript(
        f"""
        (yum download --destdir /tmp {packages[0]} {packages[1]} \
        || (dnf install -y 'dnf-command(download)' && dnf download --destdir /tmp {packages[0]} {packages[1]}) \
        || (dnf5 install -y 'dnf-command(download)' && dnf5 download --destdir /tmp {packages[0]} {packages[1]})) \
        && mv /tmp/{packages[0]}*.x86_64.rpm /tmp/{packages[0]}.rpm && mv /tmp/{packages[1]}*.x86_64.rpm /tmp/{packages[1]}.rpm
        """))  # noqa: E501

    # TODO: yum and downloaded packages results in post-install `rpm -q`
    # check to make sure packages were indeed installed - but that
    # breaks because that test uses original package paths, not package
    # names, and fails. Disable this check for now, but it's a sloppy
    # "solution".
    if package_manager_class is tmt.package_managers.dnf.Yum:
        output = package_manager.install(
            *installables,
            options=Options(
                check_first=False,
                skip_missing=True
                ))

    else:
        output = package_manager.install(
            *installables,
            options=Options(
                check_first=False
                ))

    assert_log(caplog, message=MATCH(
        rf"Run command: podman exec .+? /bin/bash -c '{expected_command}'"))

    assert_output(expected_output, output.stdout, output.stderr)


def _parametrize_test_install_debuginfo() -> Iterator[
        tuple[
            Container,
            PackageManagerClass,
            tuple[Package, Package],
            str,
            Optional[str]]]:

    for container, package_manager_class in CONTAINER_BASE_MATRIX:
        # Skip testing debuginfo install on coreos images
        if 'coreos' in container.url:
            yield pytest.param(
                container,
                package_manager_class,
                (Package('tree'), Package('dos2unix')),
                "",
                None,
                marks=pytest.mark.skip(reason="debuginfo install not supported yet on coreos")
                )

        elif package_manager_class is tmt.package_managers.dnf.Dnf5:
            yield container, \
                package_manager_class, \
                (Package('dos2unix'), Package('tree')), \
                r"debuginfo-install -y  dos2unix tree && rpm -q dos2unix-debuginfo tree-debuginfo", \
                None  # noqa: E501

        elif package_manager_class is tmt.package_managers.dnf.Dnf \
                or package_manager_class is tmt.package_managers.dnf.Yum:
            if 'centos' in container.url:
                yield pytest.param(
                    container,
                    package_manager_class,
                    (Package('dos2unix'), Package('tree')),
                    r"debuginfo-install -y  dos2unix tree && rpm -q dos2unix-debuginfo tree-debuginfo",  # noqa: E501
                    None,
                    marks=pytest.mark.skip(
                        reason='centos comes without debuginfo repos, we do not enable them yet'))

            elif 'ubi/8' in container.url:
                yield container, \
                    package_manager_class, \
                    (Package('dconf'), Package('libpng')), \
                    r"debuginfo-install -y  dconf libpng && rpm -q dconf-debuginfo libpng-debuginfo", \
                    None  # noqa: E501

            else:
                yield container, \
                    package_manager_class, \
                    (Package('dos2unix'), Package('tree')), \
                    r"debuginfo-install -y  dos2unix tree && rpm -q dos2unix-debuginfo tree-debuginfo", \
                    None  # noqa: E501

        elif package_manager_class is tmt.package_managers.apt.Apt \
                or package_manager_class is tmt.package_managers.rpm_ostree.RpmOstree \
                or package_manager_class is tmt.package_managers.apk.Apk:
            yield pytest.param(
                container,
                package_manager_class,
                (Package('tree'), Package('dos2unix')),
                "",
                None,
                marks=pytest.mark.skip(reason="not supported yet")
                )

        else:
            pytest.fail(f"Unhandled package manager class '{package_manager_class}'.")


@pytest.mark.containers
@pytest.mark.parametrize(('container_per_test',
                          'package_manager_class',
                          'installables',
                          'expected_command',
                          'expected_output'),
                         list(_parametrize_test_install_debuginfo()),
                         indirect=["container_per_test"],
                         ids=CONTAINER_MATRIX_IDS)
def test_install_debuginfo(
        container_per_test: ContainerData,
        guest_per_test: GuestContainer,
        package_manager_class: PackageManagerClass,
        installables: tuple[Package, Package],
        expected_command: str,
        expected_output: Optional[str],
        root_logger: tmt.log.Logger,
        caplog: _pytest.logging.LogCaptureFixture) -> None:

    package_manager = create_package_manager(
        container_per_test,
        guest_per_test,
        package_manager_class,
        root_logger)

    output = package_manager.install_debuginfo(*installables)

    assert_log(caplog, message=MATCH(
        rf"Run command: podman exec .+? /bin/bash -c '{expected_command}'"))

    assert_output(expected_output, output.stdout, output.stderr)


def _parametrize_test_install_debuginfo_nonexistent() -> Iterator[
        tuple[
            Container,
            PackageManagerClass,
            tuple[Package, Package],
            str,
            Optional[str]]]:

    for container, package_manager_class in CONTAINER_BASE_MATRIX:
        if package_manager_class is tmt.package_managers.dnf.Dnf5 \
                or package_manager_class is tmt.package_managers.dnf.Dnf \
                or package_manager_class is tmt.package_managers.dnf.Yum:
            yield container, \
                package_manager_class, \
                (Package('dos2unix'), Package('tree-but-spelled-wrong')), \
                r"debuginfo-install -y  dos2unix tree-but-spelled-wrong && rpm -q dos2unix-debuginfo tree-but-spelled-wrong-debuginfo", \
                None  # noqa: E501

        elif package_manager_class is tmt.package_managers.apt.Apt \
                or package_manager_class is tmt.package_managers.rpm_ostree.RpmOstree \
                or package_manager_class is tmt.package_managers.apk.Apk:
            yield pytest.param(
                container,
                package_manager_class,
                (Package('tree-but-spelled-wrong'), Package('dos2unix')),
                "",
                None,
                marks=pytest.mark.skip(reason="not supported yet")
                )

        else:
            pytest.fail(f"Unhandled package manager class '{package_manager_class}'.")


@pytest.mark.containers
@pytest.mark.parametrize(('container_per_test',
                          'package_manager_class',
                          'installables',
                          'expected_command',
                          'expected_output'),
                         list(_parametrize_test_install_debuginfo_nonexistent()),
                         indirect=["container_per_test"],
                         ids=CONTAINER_MATRIX_IDS)
def test_install_debuginfo_nonexistent(
        container_per_test: ContainerData,
        guest_per_test: GuestContainer,
        package_manager_class: PackageManagerClass,
        installables: tuple[Package, Package],
        expected_command: str,
        expected_output: Optional[str],
        root_logger: tmt.log.Logger,
        caplog: _pytest.logging.LogCaptureFixture) -> None:
    package_manager = create_package_manager(
        container_per_test,
        guest_per_test,
        package_manager_class,
        root_logger)

    with pytest.raises(tmt.utils.RunError) as excinfo:
        package_manager.install_debuginfo(*installables)

    assert_log(caplog, message=MATCH(
        rf"Run command: podman exec .+? /bin/bash -c '{expected_command}'"))

    assert excinfo.type is tmt.utils.RunError
    assert excinfo.value.returncode != 0

    assert_output(expected_output, excinfo.value.stdout, excinfo.value.stderr)


def _parametrize_test_install_debuginfo_nonexistent_skip() -> Iterator[
        tuple[
            Container,
            PackageManagerClass,
            tuple[Package, Package],
            str,
            Optional[str]]]:

    for container, package_manager_class in CONTAINER_BASE_MATRIX:
        # Skip testing debuginfo install on coreos images
        if 'coreos' in container.url:
            yield pytest.param(
                container,
                package_manager_class,
                (Package('tree'), Package('dos2unix')),
                "",
                None,
                marks=pytest.mark.skip(reason="debuginfo install not supported yet on coreos")
                )

        elif package_manager_class is tmt.package_managers.dnf.Dnf5:
            yield container, \
                package_manager_class, \
                (Package('dos2unix'), Package('tree-but-spelled-wrong')), \
                r"debuginfo-install -y --skip-broken dos2unix tree-but-spelled-wrong", \
                None

        elif package_manager_class is tmt.package_managers.dnf.Dnf \
                or package_manager_class is tmt.package_managers.dnf.Yum:
            if 'centos' in container.url:
                yield pytest.param(
                    container,
                    package_manager_class,
                    (Package('dos2unix'), Package('tree-but-spelled-wrong')),
                    r"debuginfo-install -y --skip-broken dos2unix tree-but-spelled-wrong",
                    None,
                    marks=pytest.mark.skip(
                        reason='centos comes without debuginfo repos, we do not enable them yet'))

            else:
                yield container, \
                    package_manager_class, \
                    (Package('dos2unix'), Package('tree-but-spelled-wrong')), \
                    r"debuginfo-install -y --skip-broken dos2unix tree-but-spelled-wrong", \
                    None

        elif package_manager_class is tmt.package_managers.apt.Apt \
                or package_manager_class is tmt.package_managers.rpm_ostree.RpmOstree \
                or package_manager_class is tmt.package_managers.apk.Apk:
            yield pytest.param(
                container,
                package_manager_class,
                (Package('tree-but-spelled-wrong'), Package('dos2unix')),
                "",
                None,
                marks=pytest.mark.skip(reason="not supported yet")
                )

        else:
            pytest.fail(f"Unhandled package manager class '{package_manager_class}'.")


@pytest.mark.containers
@pytest.mark.parametrize(('container_per_test',
                          'package_manager_class',
                          'installables',
                          'expected_command',
                          'expected_output'),
                         list(_parametrize_test_install_debuginfo_nonexistent_skip()),
                         indirect=["container_per_test"],
                         ids=CONTAINER_MATRIX_IDS)
def test_install_debuginfo_nonexistent_skip(
        container_per_test: ContainerData,
        guest_per_test: GuestContainer,
        package_manager_class: PackageManagerClass,
        installables: tuple[Package, Package],
        expected_command: str,
        expected_output: Optional[str],
        root_logger: tmt.log.Logger,
        caplog: _pytest.logging.LogCaptureFixture) -> None:
    package_manager = create_package_manager(
        container_per_test,
        guest_per_test,
        package_manager_class,
        root_logger)

    output = package_manager.install_debuginfo(
        *installables,
        options=Options(skip_missing=True)
        )

    assert_log(caplog, message=MATCH(
        rf"Run command: podman exec .+? /bin/bash -c '{expected_command}'"))

    assert_output(expected_output, output.stdout, output.stderr)



import asyncio
import dataclasses
import datetime
import logging
import os
from collections.abc import Mapping
from contextlib import suppress
from functools import wraps
from typing import Any, Callable, Optional, TypedDict, Union, cast

import tmt
import tmt.hardware
import tmt.log
import tmt.options
import tmt.steps
import tmt.steps.provision
import tmt.utils
from tmt.utils import (
    Command,
    Path,
    ProvisionError,
    ShellScript,
    UpdatableMessage,
    field,
    )

mrack: Any
providers: Any
ProvisioningError: Any
NotAuthenticatedError: Any
BEAKER: Any
BeakerProvider: Any
BeakerTransformer: Any
TmtBeakerTransformer: Any

_MRACK_IMPORTED: bool = False

DEFAULT_USER = 'root'
DEFAULT_ARCH = 'x86_64'
DEFAULT_IMAGE = 'fedora'
DEFAULT_PROVISION_TIMEOUT = 3600  # 1 hour timeout at least
DEFAULT_PROVISION_TICK = 60  # poll job each minute

#: How often Beaker session should be refreshed to pick up up-to-date
#: Kerberos ticket.
DEFAULT_API_SESSION_REFRESH = 3600

# Type annotation for "data" package describing a guest instance. Passed
# between load() and save() calls


class GuestInspectType(TypedDict):
    status: str
    system: str
    address: Optional[str]


# Mapping of HW requirement operators to their Beaker representation.
OPERATOR_SIGN_TO_OPERATOR = {
    tmt.hardware.Operator.EQ: '==',
    tmt.hardware.Operator.NEQ: '!=',
    tmt.hardware.Operator.GT: '>',
    tmt.hardware.Operator.GTE: '>=',
    tmt.hardware.Operator.LT: '<',
    tmt.hardware.Operator.LTE: '<=',
    }


def operator_to_beaker_op(operator: tmt.hardware.Operator, value: str) -> tuple[str, str, bool]:
    """
    Convert constraint operator to Beaker "op".

    :param operator: operator to convert.
    :param value: value operator works with. It shall be a string representation
        of the the constraint value, as converted for the Beaker job XML.
    :returns: tuple of three items: Beaker operator, fit for ``op`` attribute
        of XML filters, a value to go with it instead of the input one, and
        a boolean signalizing whether the filter, constructed by the caller,
        should be negated.
    """

    if operator in OPERATOR_SIGN_TO_OPERATOR:
        return OPERATOR_SIGN_TO_OPERATOR[operator], value, False

    # MATCH has special handling - convert the pattern to a wildcard form -
    # and that may be weird :/
    if operator == tmt.hardware.Operator.MATCH:
        return 'like', value.replace('.*', '%').replace('.+', '%'), False

    if operator == tmt.hardware.Operator.NOTMATCH:
        return 'like', value.replace('.*', '%').replace('.+', '%'), True

    raise ProvisionError(f"Hardware requirement operator '{operator}' is not supported.")


# Transcription of our HW constraints into Mrack's own representation. It's based
# on dictionaries, and it's slightly weird. There is no distinction between elements
# that do not have attributes, like <and/>, and elements that must have them, like
# <memory/> and other binary operations. Also, there is no distinction between
# element attribute and child element, both are specified as dictionary key, just
# the former would be a string, the latter another, nested, dictionary.
#
# This makes it harder for us to enforce correct structure of the transcribed tree.
# Therefore adding a thin layer of containers that describe what Mrack is willing
# to accept, but with strict type annotations; the layer is aware of how to convert
# its components into dictionaries.
@dataclasses.dataclass
class MrackBaseHWElement:
    """ Base for Mrack hardware requirement elements """

    # Only a name is defined, as it's the only property shared across all element
    # types.
    name: str

    def to_mrack(self) -> dict[str, Any]:
        """ Convert the element to Mrack-compatible dictionary tree """
        raise NotImplementedError


@dataclasses.dataclass
class MrackHWElement(MrackBaseHWElement):
    """
    An element with name and attributes.

    This type of element is not allowed to have any child elements.
    """

    attributes: dict[str, str] = dataclasses.field(default_factory=dict)

    def to_mrack(self) -> dict[str, Any]:
        return {
            self.name: self.attributes
            }


@dataclasses.dataclass(init=False)
class MrackHWBinOp(MrackHWElement):
    """ An element describing a binary operation, a "check" """

    def __init__(self, name: str, operator: str, value: str) -> None:
        super().__init__(name)

        self.attributes = {
            '_op': operator,
            '_value': value
            }


@dataclasses.dataclass(init=False)
class MrackHWKeyValue(MrackHWElement):
    """ A key-value element """

    def __init__(self, name: str, operator: str, value: str) -> None:
        super().__init__('key_value')

        self.attributes = {
            '_key': name,
            '_op': operator,
            '_value': value
            }


@dataclasses.dataclass
class MrackHWGroup(MrackBaseHWElement):
    """
    An element with child elements.

    This type of element is not allowed to have any attributes.
    """

    children: list[MrackBaseHWElement] = dataclasses.field(default_factory=list)

    def to_mrack(self) -> dict[str, Any]:
        # Another unexpected behavior of mrack dictionary tree: if there is just
        # a single child, it is "packed" into its parent as a key/dict item.
        if len(self.children) == 1 and self.name not in ('and', 'or'):
            return {
                self.name: self.children[0].to_mrack()
                }

        return {
            self.name: [child.to_mrack() for child in self.children]
            }


@dataclasses.dataclass
class MrackHWAndGroup(MrackHWGroup):
    """ Represents ``<and/>`` element """

    name: str = 'and'


@dataclasses.dataclass
class MrackHWOrGroup(MrackHWGroup):
    """ Represents ``<or/>`` element """

    name: str = 'or'


@dataclasses.dataclass
class MrackHWNotGroup(MrackHWGroup):
    """ Represents ``<not/>`` element """

    name: str = 'not'


def _transform_unsupported(
        constraint: tmt.hardware.Constraint[Any],
        logger: tmt.log.Logger) -> MrackBaseHWElement:
    # Unsupported constraint has been already logged via report_support(). Make
    # sure user is aware it would have no effect, and since we have to return
    # something, return an empty `or` group - no harm done, composable with other
    # elements.
    logger.warning(f"Hardware requirement '{constraint.printable_name}' will have no effect.")

    return MrackHWOrGroup()


def _transform_beaker_pool(
        constraint: tmt.hardware.TextConstraint,
        logger: tmt.log.Logger) -> MrackBaseHWElement:
    beaker_operator, actual_value, negate = operator_to_beaker_op(
        constraint.operator,
        constraint.value)

    return MrackHWBinOp(
        'pool',
        beaker_operator,
        actual_value)


def _transform_cpu_flag(
        constraint: tmt.hardware.TextConstraint,
        logger: tmt.log.Logger) -> MrackBaseHWElement:
    beaker_operator = OPERATOR_SIGN_TO_OPERATOR[tmt.hardware.Operator.EQ] \
        if constraint.operator is tmt.hardware.Operator.CONTAINS \
        else OPERATOR_SIGN_TO_OPERATOR[tmt.hardware.Operator.NEQ]
    actual_value = str(constraint.value)

    return MrackHWGroup(
        'cpu',
        children=[MrackHWBinOp('flag', beaker_operator, actual_value)]
        )


def _transform_cpu_model(
        constraint: tmt.hardware.IntegerConstraint,
        logger: tmt.log.Logger) -> MrackBaseHWElement:
    beaker_operator, actual_value, _ = operator_to_beaker_op(
        constraint.operator,
        str(constraint.value))

    return MrackHWGroup(
        'cpu',
        children=[MrackHWBinOp('model', beaker_operator, actual_value)])


def _transform_cpu_processors(
        constraint: tmt.hardware.IntegerConstraint,
        logger: tmt.log.Logger) -> MrackBaseHWElement:
    beaker_operator, actual_value, _ = operator_to_beaker_op(
        constraint.operator,
        str(constraint.value))

    return MrackHWGroup(
        'cpu',
        children=[MrackHWBinOp('processors', beaker_operator, actual_value)])


def _transform_cpu_cores(
        constraint: tmt.hardware.IntegerConstraint,
        logger: tmt.log.Logger) -> MrackBaseHWElement:
    beaker_operator, actual_value, _ = operator_to_beaker_op(
        constraint.operator,
        str(constraint.value))

    return MrackHWGroup(
        'cpu',
        children=[MrackHWBinOp('cores', beaker_operator, actual_value)])


def _transform_cpu_model_name(
        constraint: tmt.hardware.TextConstraint,
        logger: tmt.log.Logger) -> MrackBaseHWElement:
    beaker_operator, actual_value, negate = operator_to_beaker_op(
        constraint.operator,
        constraint.value)

    if negate:
        return MrackHWNotGroup(children=[
            MrackHWGroup(
                'cpu',
                children=[MrackHWBinOp('model_name', beaker_operator, actual_value)])])

    return MrackHWGroup(
        'cpu',
        children=[MrackHWBinOp('model_name', beaker_operator, actual_value)])


def _transform_cpu_stepping(
        constraint: tmt.hardware.IntegerConstraint,
        logger: tmt.log.Logger) -> MrackBaseHWElement:
    beaker_operator, actual_value, _ = operator_to_beaker_op(
        constraint.operator,
        str(constraint.value))

    return MrackHWGroup(
        'cpu',
        children=[MrackHWBinOp('stepping', beaker_operator, actual_value)])


def _transform_cpu_vendor_name(
        constraint: tmt.hardware.TextConstraint,
        logger: tmt.log.Logger) -> MrackBaseHWElement:
    beaker_operator, actual_value, negate = operator_to_beaker_op(
        constraint.operator,
        str(constraint.value))

    if negate:
        return MrackHWNotGroup(children=[
            MrackHWGroup(
                'cpu',
                children=[MrackHWBinOp('vendor', beaker_operator, actual_value)])])

    return MrackHWGroup(
        'cpu',
        children=[MrackHWBinOp('vendor', beaker_operator, actual_value)])


def _transform_cpu_hyper_threading(
        constraint: tmt.hardware.FlagConstraint,
        logger: tmt.log.Logger) -> MrackBaseHWElement:
    beaker_operator, actual_value, _ = operator_to_beaker_op(
        constraint.operator,
        str(constraint.value))

    return MrackHWGroup(
        'cpu',
        children=[MrackHWBinOp('hyper', beaker_operator, actual_value)])


def _transform_disk_driver(
        constraint: tmt.hardware.TextConstraint,
        logger: tmt.log.Logger) -> MrackBaseHWElement:
    beaker_operator, actual_value, negate = operator_to_beaker_op(
        constraint.operator,
        constraint.value)

    if negate:
        return MrackHWNotGroup(children=[
            MrackHWKeyValue('BOOTDISK', beaker_operator, actual_value)
            ])

    return MrackHWKeyValue(
        'BOOTDISK',
        beaker_operator,
        actual_value)


def _transform_disk_size(
        constraint: tmt.hardware.SizeConstraint,
        logger: tmt.log.Logger) -> MrackBaseHWElement:
    beaker_operator, actual_value, _ = operator_to_beaker_op(
        constraint.operator,
        str(int(constraint.value.to('B').magnitude))
        )

    return MrackHWGroup(
        'disk',
        children=[MrackHWBinOp('size', beaker_operator, actual_value)])


def _transform_disk_model_name(
        constraint: tmt.hardware.TextConstraint,
        logger: tmt.log.Logger) -> MrackBaseHWElement:
    beaker_operator, actual_value, negate = operator_to_beaker_op(
        constraint.operator,
        constraint.value)

    if negate:
        return MrackHWNotGroup(children=[
            MrackHWGroup(
                'disk',
                children=[MrackHWBinOp('model', beaker_operator, actual_value)])])

    return MrackHWGroup(
        'disk',
        children=[MrackHWBinOp('model', beaker_operator, actual_value)])


def _transform_disk_physical_sector_size(
        constraint: tmt.hardware.SizeConstraint,
        logger: tmt.log.Logger) -> MrackBaseHWElement:
    beaker_operator, actual_value, _ = operator_to_beaker_op(
        constraint.operator,
        str(constraint.value)
        )

    return MrackHWGroup(
        'disk',
        children=[MrackHWBinOp('phys_sector_size', beaker_operator, actual_value)])


def _transform_disk_logical_sector_size(
        constraint: tmt.hardware.SizeConstraint,
        logger: tmt.log.Logger) -> MrackBaseHWElement:
    beaker_operator, actual_value, _ = operator_to_beaker_op(
        constraint.operator,
        str(constraint.value)
        )

    return MrackHWGroup(
        'disk',
        children=[MrackHWBinOp('sector_size', beaker_operator, actual_value)])


def _transform_hostname(
        constraint: tmt.hardware.TextConstraint,
        logger: tmt.log.Logger) -> MrackBaseHWElement:
    beaker_operator, actual_value, negate = operator_to_beaker_op(
        constraint.operator,
        constraint.value)

    if negate:
        return MrackHWNotGroup(children=[
            MrackHWBinOp('hostname', beaker_operator, actual_value)
            ])

    return MrackHWBinOp(
        'hostname',
        beaker_operator,
        actual_value)


def _transform_memory(
        constraint: tmt.hardware.SizeConstraint,
        logger: tmt.log.Logger) -> MrackBaseHWElement:

    beaker_operator, actual_value, _ = operator_to_beaker_op(
        constraint.operator,
        str(int(constraint.value.to('MiB').magnitude)))

    return MrackHWGroup(
        'system',
        children=[MrackHWBinOp('memory', beaker_operator, actual_value)])


def _transform_tpm_version(
        constraint: tmt.hardware.TextConstraint,
        logger: tmt.log.Logger) -> MrackBaseHWElement:
    beaker_operator, actual_value, negate = operator_to_beaker_op(
        constraint.operator,
        constraint.value)

    return MrackHWKeyValue('TPM', beaker_operator, actual_value)


def _transform_virtualization_is_virtualized(
        constraint: tmt.hardware.FlagConstraint,
        logger: tmt.log.Logger) -> MrackBaseHWElement:
    beaker_operator, actual_value, _ = operator_to_beaker_op(
        constraint.operator,
        str(constraint.value))

    test = (constraint.operator, constraint.value)

    if test in [(tmt.hardware.Operator.EQ, True), (tmt.hardware.Operator.NEQ, False)]:
        return MrackHWGroup(
            'system',
            children=[MrackHWBinOp('hypervisor', '!=', '')])

    if test in [(tmt.hardware.Operator.EQ, False), (tmt.hardware.Operator.NEQ, True)]:
        return MrackHWGroup(
            'system',
            children=[MrackHWBinOp('hypervisor', '==', '')])

    return _transform_unsupported(constraint, logger)


def _transform_virtualization_hypervisor(
        constraint: tmt.hardware.TextConstraint,
        logger: tmt.log.Logger) -> MrackBaseHWElement:
    beaker_operator, actual_value, negate = operator_to_beaker_op(
        constraint.operator,
        str(constraint.value))

    if negate:
        return MrackHWNotGroup(children=[
            MrackHWGroup(
                'system',
                children=[MrackHWBinOp('hypervisor', beaker_operator, actual_value)])])

    return MrackHWGroup(
        'system',
        children=[MrackHWBinOp('hypervisor', beaker_operator, actual_value)])


def _transform_zcrypt_adapter(
        constraint: tmt.hardware.TextConstraint,
        logger: tmt.log.Logger) -> MrackBaseHWElement:
    beaker_operator, actual_value, negate = operator_to_beaker_op(
        constraint.operator,
        constraint.value)

    if negate:
        return MrackHWNotGroup(children=[
            MrackHWGroup(
                'system',
                children=[MrackHWKeyValue('ZCRYPT_MODEL', beaker_operator, actual_value)])])

    return MrackHWGroup(
        'system',
        children=[MrackHWKeyValue('ZCRYPT_MODEL', beaker_operator, actual_value)])


def _transform_zcrypt_mode(
        constraint: tmt.hardware.TextConstraint,
        logger: tmt.log.Logger) -> MrackBaseHWElement:
    beaker_operator, actual_value, negate = operator_to_beaker_op(
        constraint.operator,
        constraint.value)

    if negate:
        return MrackHWNotGroup(children=[
            MrackHWGroup(
                'system',
                children=[MrackHWKeyValue('ZCRYPT_MODE', beaker_operator, actual_value)])])

    return MrackHWGroup(
        'system',
        children=[MrackHWKeyValue('ZCRYPT_MODE', beaker_operator, actual_value)])


def _transform_iommu_is_supported(
        constraint: tmt.hardware.FlagConstraint,
        logger: tmt.log.Logger) -> MrackBaseHWElement:

    test = (constraint.operator, constraint.value)

    if test in [(tmt.hardware.Operator.EQ, True), (tmt.hardware.Operator.NEQ, False)]:
        return MrackHWKeyValue('VIRT_IOMMU', '==', '1')

    if test in [(tmt.hardware.Operator.EQ, False), (tmt.hardware.Operator.NEQ, True)]:
        return MrackHWKeyValue('VIRT_IOMMU', '==', '0')

    return _transform_unsupported(constraint, logger)


def _transform_location_lab_controller(
        constraint: tmt.hardware.TextConstraint,
        logger: tmt.log.Logger) -> MrackBaseHWElement:
    if constraint.operator not in [tmt.hardware.Operator.EQ, tmt.hardware.Operator.NEQ]:
        raise ProvisionError(
            f"Cannot apply hardware requirement '{constraint}', operator not supported.")
    beaker_operator, actual_value, negate = operator_to_beaker_op(
        constraint.operator,
        constraint.value)

    if negate:
        return MrackHWNotGroup(children=[
            MrackHWBinOp('labcontroller', beaker_operator, actual_value)
            ])

    return MrackHWBinOp(
        'labcontroller',
        beaker_operator,
        actual_value)


def _transform_system_numa_nodes(
        constraint: tmt.hardware.IntegerConstraint,
        logger: tmt.log.Logger) -> MrackBaseHWElement:
    beaker_operator, actual_value, _ = operator_to_beaker_op(
        constraint.operator,
        str(constraint.value))

    return MrackHWGroup(
        'system',
        children=[MrackHWBinOp('numanodes', beaker_operator, actual_value)])


ConstraintTransformer = Callable[[
    tmt.hardware.Constraint[Any], tmt.log.Logger], MrackBaseHWElement]

_CONSTRAINT_TRANSFORMERS: Mapping[str, ConstraintTransformer] = {
    'beaker.pool': _transform_beaker_pool,  # type: ignore[dict-item]
    'cpu.cores': _transform_cpu_cores,  # type: ignore[dict-item]
    'cpu.flag': _transform_cpu_flag,  # type: ignore[dict-item]
    'cpu.hyper_threading': _transform_cpu_hyper_threading,  # type: ignore[dict-item]
    'cpu.model': _transform_cpu_model,  # type: ignore[dict-item]
    'cpu.model_name': _transform_cpu_model_name,  # type: ignore[dict-item]
    'cpu.processors': _transform_cpu_processors,  # type: ignore[dict-item]
    'cpu.stepping': _transform_cpu_stepping,  # type: ignore[dict-item]
    'cpu.vendor_name': _transform_cpu_vendor_name,  # type: ignore[dict-item]
    'disk.driver': _transform_disk_driver,  # type: ignore[dict-item]
    'disk.model_name': _transform_disk_model_name,  # type: ignore[dict-item]
    'disk.size': _transform_disk_size,  # type: ignore[dict-item]
    'disk.physical_sector_size': _transform_disk_physical_sector_size,  # type: ignore[dict-item]
    'disk.logical_sector_size': _transform_disk_logical_sector_size,  # type: ignore[dict-item]
    'hostname': _transform_hostname,  # type: ignore[dict-item]
    'location.lab_controller': _transform_location_lab_controller,  # type: ignore[dict-item]
    'memory': _transform_memory,  # type: ignore[dict-item]
    'tpm.version': _transform_tpm_version,  # type: ignore[dict-item]
    'virtualization.is_virtualized':
        _transform_virtualization_is_virtualized,  # type: ignore[dict-item]
    'virtualization.hypervisor':
        _transform_virtualization_hypervisor,  # type: ignore[dict-item]
    'zcrypt.adapter': _transform_zcrypt_adapter,  # type: ignore[dict-item]
    'zcrypt.mode': _transform_zcrypt_mode,  # type: ignore[dict-item]
    'system.numa_nodes': _transform_system_numa_nodes,  # type: ignore[dict-item]
    'iommu.is_supported': _transform_iommu_is_supported,  # type: ignore[dict-item]
    }


def constraint_to_beaker_filter(
        constraint: tmt.hardware.BaseConstraint,
        logger: tmt.log.Logger) -> MrackBaseHWElement:
    """ Convert a hardware constraint into a Mrack-compatible filter """

    if isinstance(constraint, tmt.hardware.And):
        return MrackHWAndGroup(
            children=[
                constraint_to_beaker_filter(child_constraint, logger)
                for child_constraint in constraint.constraints
                ]
            )

    if isinstance(constraint, tmt.hardware.Or):
        return MrackHWOrGroup(
            children=[
                constraint_to_beaker_filter(child_constraint, logger)
                for child_constraint in constraint.constraints
                ]
            )

    assert isinstance(constraint, tmt.hardware.Constraint)

    name, _, child_name = constraint.expand_name()

    if child_name:
        transformer = _CONSTRAINT_TRANSFORMERS.get(f'{name}.{child_name}')

    else:
        transformer = _CONSTRAINT_TRANSFORMERS.get(name)

    if transformer:
        return transformer(constraint, logger)

    return _transform_unsupported(constraint, logger)


def import_and_load_mrack_deps(workdir: Any, name: str, logger: tmt.log.Logger) -> None:
    """ Import mrack module only when needed """
    global _MRACK_IMPORTED

    if _MRACK_IMPORTED:
        return

    global mrack
    global providers
    global ProvisioningError
    global NotAuthenticatedError
    global BEAKER
    global BeakerProvider
    global BeakerTransformer
    global TmtBeakerTransformer

    try:
        import mrack
        from mrack.errors import NotAuthenticatedError, ProvisioningError
        from mrack.providers import providers
        from mrack.providers.beaker import PROVISIONER_KEY as BEAKER
        from mrack.providers.beaker import BeakerProvider
        from mrack.transformers.beaker import BeakerTransformer

        # hack: remove mrack stdout and move the logfile to /tmp
        mrack.logger.removeHandler(mrack.console_handler)
        mrack.logger.removeHandler(mrack.file_handler)

        with suppress(OSError):
            os.remove("mrack.log")

        logging.FileHandler(str(f"{workdir}/{name}-mrack.log"))

        providers.register(BEAKER, BeakerProvider)

    except ImportError:
        raise ProvisionError(
            "Install 'tmt+provision-beaker' to provision using this method.")

    # ignore the misc because mrack sources are not typed and result into
    # error: Class cannot subclass "BeakerTransformer" (has type "Any")
    # as mypy does not have type information for the BeakerTransformer class
    class TmtBeakerTransformer(BeakerTransformer):  # type: ignore[misc]
        def _translate_tmt_hw(self, hw: tmt.hardware.Hardware) -> dict[str, Any]:
            """ Return hw requirements from given hw dictionary """

            assert hw.constraint

            transformed = MrackHWAndGroup(
                children=[
                    constraint_to_beaker_filter(constraint, logger)
                    for constraint in hw.constraint.variant()
                    ])

            logger.debug('Transformed hardware', tmt.utils.dict_to_yaml(transformed.to_mrack()))

            return {
                'hostRequires': transformed.to_mrack()
                }

        def create_host_requirement(self, host: CreateJobParameters) -> dict[str, Any]:
            """ Create single input for Beaker provisioner """
            req: dict[str, Any] = super().create_host_requirement(host.to_mrack())

            if host.hardware and host.hardware.constraint:
                req.update(self._translate_tmt_hw(host.hardware))

            if host.beaker_job_owner:
                req['job_owner'] = host.beaker_job_owner

            # Whiteboard must be added *after* request preparation, to overwrite the default one.
            req['whiteboard'] = host.whiteboard

            logger.debug('mrack request', req, level=4)

            logger.info('whiteboard', host.whiteboard, 'green')

            return req

    _MRACK_IMPORTED = True


def async_run(func: Any) -> Any:
    """ Decorate click actions to run as async """
    @wraps(func)
    def update_wrapper(*args: Any, **kwargs: Any) -> Any:
        return asyncio.run(func(*args, **kwargs))

    return update_wrapper


@dataclasses.dataclass
class BeakerGuestData(tmt.steps.provision.GuestSshData):
    # Override parent class with our defaults
    user: str = field(
        default=DEFAULT_USER,
        option=('-u', '--user'),
        metavar='USERNAME',
        help='Username to use for all guest operations.')

    # Guest request properties
    whiteboard: Optional[str] = field(
        default=None,
        option=('-w', '--whiteboard'),
        metavar='WHITEBOARD',
        help='Text description of the beaker job which is displayed in the list of jobs.'
        )
    arch: str = field(
        default=DEFAULT_ARCH,
        option='--arch',
        metavar='ARCH',
        help='Architecture to provision.')
    image: Optional[str] = field(
        default=DEFAULT_IMAGE,
        option=('-i', '--image'),
        metavar='COMPOSE',
        help='Image (distro or "compose" in Beaker terminology) to provision.')

    # Provided in Beaker job
    job_id: Optional[str] = field(
        default=None,
        internal=True)

    # Timeouts and deadlines
    provision_timeout: int = field(
        default=DEFAULT_PROVISION_TIMEOUT,
        option='--provision-timeout',
        metavar='SECONDS',
        help=f"""
             How long to wait for provisioning to complete,
             {DEFAULT_PROVISION_TIMEOUT} seconds by default.
             """,
        normalize=tmt.utils.normalize_int)
    provision_tick: int = field(
        default=DEFAULT_PROVISION_TICK,
        option='--provision-tick',
        metavar='SECONDS',
        help=f"""
             How often check Beaker for provisioning status,
             {DEFAULT_PROVISION_TICK} seconds by default.
             """,
        normalize=tmt.utils.normalize_int)
    api_session_refresh_tick: int = field(
        default=DEFAULT_API_SESSION_REFRESH,
        option='--api-session-refresh-tick',
        metavar='SECONDS',
        help=f"""
             How often should Beaker session be refreshed to pick up-to-date Kerberos ticket,
             {DEFAULT_API_SESSION_REFRESH} seconds by default.
             """,
        normalize=tmt.utils.normalize_int)
    kickstart: dict[str, str] = field(
        default_factory=dict,
        option='--kickstart',
        metavar='KEY=VALUE',
        help='Optional Beaker kickstart to use when provisioning the guest.',
        multiple=True,
        normalize=tmt.utils.normalize_string_dict)

    beaker_job_owner: Optional[str] = field(
        default=None,
        option='--beaker-job-owner',
        metavar='USERNAME',
        help="""
             If set, Beaker jobs will be submitted on behalf of ``USERNAME``.
             Submitting user must be a submission delegate for the ``USERNAME``.
             """)


@dataclasses.dataclass
class ProvisionBeakerData(BeakerGuestData, tmt.steps.provision.ProvisionStepData):
    pass


GUEST_STATE_COLOR_DEFAULT = 'green'

GUEST_STATE_COLORS = {
    "New": "blue",
    "Scheduled": "blue",
    "Queued": "cyan",
    "Processed": "cyan",
    "Waiting": "magenta",
    "Installing": "magenta",
    "Running": "magenta",
    "Cancelled": "yellow",
    "Aborted": "yellow",
    "Reserved": "green",
    "Completed": "green",
    }


@dataclasses.dataclass
class CreateJobParameters:
    """ Collect all parameters for a future Beaker job """

    tmt_name: str
    name: str
    os: str
    arch: str
    hardware: Optional[tmt.hardware.Hardware]
    kickstart: dict[str, str]
    whiteboard: Optional[str]
    beaker_job_owner: Optional[str]
    group: str = 'linux'

    def to_mrack(self) -> dict[str, Any]:
        data = dataclasses.asdict(self)

        data['beaker'] = {}

        if self.kickstart:
            data['beaker']['ks_meta'] = self.kickstart.get('metadata')
            data['beaker']['ks_append'] = self.kickstart

        return data


class BeakerAPI:
    # req is a requirement passed to Beaker mrack provisioner
    mrack_requirement: dict[str, Any] = {}
    dsp_name: str = "Beaker"

    # wrapping around the __init__ with async wrapper does mangle the method
    # and mypy complains as it no longer returns None but the coroutine
    @async_run
    async def __init__(self, guest: 'GuestBeaker') -> None:  # type: ignore[misc]
        """ Initialize the API class with defaults and load the config """
        self._guest = guest

        # use global context class
        global_context = mrack.context.global_context

        mrack_config_locations = [
            Path(__file__).parent / "mrack/mrack.conf",
            Path("/etc/tmt/mrack.conf"),
            Path("~/.mrack/mrack.conf").expanduser(),
            Path.cwd() / "mrack.conf"
            ]

        mrack_config: Optional[Path] = None

        for potential_location in mrack_config_locations:
            if potential_location.exists():
                mrack_config = potential_location

        if not mrack_config:
            raise ProvisionError("Configuration file 'mrack.conf' not found.")

        try:
            global_context.init(str(mrack_config))
        except mrack.errors.ConfigError as mrack_conf_err:
            raise ProvisionError(mrack_conf_err)

        self._mrack_transformer = TmtBeakerTransformer()
        try:
            await self._mrack_transformer.init(global_context.PROV_CONFIG, {})
        except NotAuthenticatedError as kinit_err:
            raise ProvisionError(kinit_err) from kinit_err
        except AttributeError as hub_err:
            raise ProvisionError(
                f"Can not use current kerberos ticket to authenticate: {hub_err}"
                ) from hub_err
        except FileNotFoundError as missing_conf_err:
            raise ProvisionError(
                f"Configuration file missing: {missing_conf_err.filename}"
                ) from missing_conf_err

        self._mrack_provider = self._mrack_transformer._provider
        self._mrack_provider.poll_sleep = DEFAULT_PROVISION_TICK

        if guest.job_id:
            self._bkr_job_id = guest.job_id

    @async_run
    async def create(
            self,
            data: CreateJobParameters) -> Any:
        """
        Create - or request creation of - a resource using mrack up.

        :param data: describes the provisioning request.
        """

        mrack_requirement = self._mrack_transformer.create_host_requirement(data)
        log_msg_start = f"{self.dsp_name} [{self.mrack_requirement.get('name')}]"
        self._bkr_job_id, self._req = await self._mrack_provider.create_server(mrack_requirement)
        return self._mrack_provider._get_recipe_info(self._bkr_job_id, log_msg_start)

    @async_run
    async def inspect(
            self,
            ) -> Any:
        """ Inspect a resource (kinda wait till provisioned) """
        log_msg_start = f"{self.dsp_name} [{self.mrack_requirement.get('name')}]"
        return self._mrack_provider._get_recipe_info(self._bkr_job_id, log_msg_start)

    @async_run
    async def delete(  # destroy
            self,
            ) -> Any:
        """ Delete - or request removal of - a resource """
        return await self._mrack_provider.delete_host(self._bkr_job_id, None)


class GuestBeaker(tmt.steps.provision.GuestSsh):
    """ Beaker guest instance """

    _data_class = BeakerGuestData

    # Guest request properties
    whiteboard: Optional[str]
    arch: str
    image: str = "fedora-latest"
    hardware: Optional[tmt.hardware.Hardware] = None
    kickstart: dict[str, str]

    beaker_job_owner: Optional[str] = None

    # Provided in Beaker response
    job_id: Optional[str]

    # Timeouts and deadlines
    provision_timeout: int
    provision_tick: int
    api_session_refresh_tick: int

    _api: Optional[BeakerAPI] = None
    _api_timestamp: Optional[datetime.datetime] = None

    @property
    def api(self) -> BeakerAPI:
        """ Create BeakerAPI leveraging mrack """

        def _construct_api() -> tuple[BeakerAPI, datetime.datetime]:
            assert self.parent is not None

            import_and_load_mrack_deps(self.parent.workdir, self.parent.name, self._logger)

            return BeakerAPI(self), datetime.datetime.now(datetime.timezone.utc)

        if self._api is None:
            self._api, self._api_timestamp = _construct_api()

        else:
            assert self._api_timestamp is not None

            delta = datetime.datetime.now(datetime.timezone.utc) - self._api_timestamp

            if delta.total_seconds() >= self.api_session_refresh_tick:
                self.debug(f'Refresh Beaker API client as it is too old, {delta}.')

                self._api, self._api_timestamp = _construct_api()

        return self._api

    @property
    def is_ready(self) -> bool:
        """ Check if provisioning of machine is done """
        if self.job_id is None:
            return False

        assert mrack is not None

        try:
            response = self.api.inspect()

            if response["status"] == "Aborted":
                return False

            current = cast(GuestInspectType, response)
            state = current["status"]
            if state in {"Error, Aborted", "Cancelled"}:
                return False

            if state == 'Reserved':
                return True
            return False

        except mrack.errors.MrackError:
            return False

    def _create(self, tmt_name: str) -> None:
        """ Create beaker job xml request and submit it to Beaker hub """

        data = CreateJobParameters(
            tmt_name=tmt_name,
            hardware=self.hardware,
            kickstart=self.kickstart,
            arch=self.arch,
            os=self.image,
            name=f'{self.image}-{self.arch}',
            whiteboard=self.whiteboard or tmt_name,
            beaker_job_owner=self.beaker_job_owner)

        try:
            response = self.api.create(data)

        except ProvisioningError as exc:
            import xmlrpc.client

            cause = exc.__cause__

            if isinstance(cause, xmlrpc.client.Fault):
                if 'is not a valid user name' in cause.faultString:
                    raise ProvisionError(
                        f"Failed to create Beaker job, job owner '{self.beaker_job_owner}' "
                        "was refused as unknown.") from exc

                if 'is not a valid submission delegate' in cause.faultString:
                    raise ProvisionError(
                        f"Failed to create Beaker job, job owner '{self.beaker_job_owner}' "
                        "is not a valid submission delegate.") from exc

            raise ProvisionError('Failed to create Beaker job') from exc

        if response:
            self.info('guest', 'has been requested', 'green')

        else:
            raise ProvisionError(
                f"Failed to create, response: '{response}'.")

        self.job_id = f'J:{response["id"]}'
        self.info('job id', self.job_id, 'green')

        with UpdatableMessage("status", indent_level=self._level()) as progress_message:

            def get_new_state() -> GuestInspectType:
                response = self.api.inspect()

                if response["status"] == "Aborted":
                    raise ProvisionError(
                        f"Failed to create, "
                        f"unhandled API response '{response['status']}'."
                        )

                current = cast(GuestInspectType, response)
                state = current["status"]
                state_color = GUEST_STATE_COLORS.get(
                    state, GUEST_STATE_COLOR_DEFAULT
                    )

                progress_message.update(state, color=state_color)

                if state in {"Error, Aborted", "Cancelled"}:
                    raise ProvisionError(
                        'Failed to create, provisioning failed.'
                        )

                if state == 'Reserved':
                    return current

                raise tmt.utils.WaitingIncompleteError

            try:
                guest_info = tmt.utils.wait(
                    self, get_new_state, datetime.timedelta(
                        seconds=self.provision_timeout),
                    tick=self.provision_tick
                    )

            except tmt.utils.WaitingTimedOutError:
                response = self.api.delete()
                raise ProvisionError(
                    f'Failed to provision in the given amount '
                    f'of time (--provision-timeout={self.provision_timeout}).'
                    )

        self.primary_address = self.topology_address = guest_info['system']

    def start(self) -> None:
        """
        Start the guest

        Get a new guest instance running. This should include preparing
        any configuration necessary to get it started. Called after
        load() is completed so all guest data should be available.
        """

        if self.job_id is None or self.primary_address is None:
            self._create(self._tmt_name())

        self.verbose('primary address', self.primary_address, 'green')
        self.verbose('topology address', self.topology_address, 'green')

    def stop(self) -> None:
        """ Stop the guest """
        # do nothing
        return

    def remove(self) -> None:
        """ Remove the guest """

        if self.job_id is None:
            return

        self.api.delete()

    def reboot(
            self,
            hard: bool = False,
            command: Optional[Union[Command, ShellScript]] = None,
            timeout: Optional[int] = None,
            tick: float = tmt.utils.DEFAULT_WAIT_TICK,
            tick_increase: float = tmt.utils.DEFAULT_WAIT_TICK_INCREASE) -> bool:
        """
        Reboot the guest, and wait for the guest to recover.

        :param hard: if set, force the reboot. This may result in a loss of
            data. The default of ``False`` will attempt a graceful reboot.
        :param command: a command to run on the guest to trigger the reboot.
            If not set, plugin would try to use ``bkr system-power`` for hard
            reboot. Unlike ``command``, this would be executed on the runner,
            **not** on the guest.
        :param timeout: amount of time in which the guest must become available
            again.
        :param tick: how many seconds to wait between two consecutive attempts
            of contacting the guest.
        :param tick_increase: a multiplier applied to ``tick`` after every
            attempt.
        :returns: ``True`` if the reboot succeeded, ``False`` otherwise.
        """

        if not command and hard:
            self.debug("Reboot using the reboot command 'bkr system-power --action reboot'.")

            reboot_script = ShellScript(f'bkr system-power --action reboot {self.primary_address}')

            return self.perform_reboot(
                lambda: self._run_guest_command(reboot_script.to_shell_command()),
                timeout=timeout,
                tick=tick,
                tick_increase=tick_increase,
                hard=True)

        return super().reboot(
            hard=hard,
            command=command,
            timeout=timeout,
            tick=tick,
            tick_increase=tick_increase)


@tmt.steps.provides_method('beaker')
class ProvisionBeaker(tmt.steps.provision.ProvisionPlugin[ProvisionBeakerData]):
    """
    Provision guest on Beaker system using mrack.

    Minimal configuration could look like this:

    .. code-block:: yaml

        provision:
            how: beaker
            image: fedora

    """

    _data_class = ProvisionBeakerData
    _guest_class = GuestBeaker

    # _thread_safe = True

    # Guest instance
    _guest = None

    # data argument should be a "Optional[GuestData]" type but we would like to use
    # BeakerGuestData created here ignoring the override will make mypy calm
    def wake(self, data: Optional[BeakerGuestData] = None) -> None:  # type: ignore[override]
        """ Wake up the plugin, process data, apply options """
        super().wake(data=data)

        if data:
            self._guest = GuestBeaker(
                data=data,
                name=self.name,
                parent=self.step,
                logger=self._logger,
                )

    def go(self, *, logger: Optional[tmt.log.Logger] = None) -> None:
        """ Provision the guest """
        super().go(logger=logger)

        data = BeakerGuestData.from_plugin(self)

        data.show(verbose=self.verbosity_level, logger=self._logger)

        if data.hardware:
            data.hardware.report_support(
                names=list(_CONSTRAINT_TRANSFORMERS.keys()),
                logger=self._logger)

        self._guest = GuestBeaker(
            data=data,
            name=self.name,
            parent=self.step,
            logger=self._logger,
            )
        self._guest.start()
        self._guest.setup()

    def guest(self) -> Optional[GuestBeaker]:
        """ Return the provisioned guest """
        return self._guest



import dataclasses
import webbrowser
from typing import Optional

from jinja2 import select_autoescape

import tmt
import tmt.log
import tmt.options
import tmt.steps
import tmt.steps.report
import tmt.utils
import tmt.utils.templates
from tmt.utils import Path, field

HTML_TEMPLATE_PATH = tmt.utils.resource_files('steps/report/html/template.html.j2')


@dataclasses.dataclass
class ReportHtmlData(tmt.steps.report.ReportStepData):
    open: bool = field(
        default=False,
        option=('-o', '--open'),
        is_flag=True,
        help='Open results in your preferred web browser.'
        )

    absolute_paths: bool = field(
        default=False,
        option='--absolute-paths',
        is_flag=True,
        help='Make paths absolute rather than relative to working directory.'
        )

    display_guest: str = field(
        default='auto',
        option='--display-guest',
        metavar='auto|always|never',
        choices=['auto', 'always', 'never'],
        help="""
             When to display full guest name in report: when more than a single guest was involved
             (default), always, or never.
             """)


@tmt.steps.provides_method('html')
class ReportHtml(tmt.steps.report.ReportPlugin[ReportHtmlData]):
    """
    Format test results into an HTML report.

    Example config:

    .. code-block:: yaml

        report:
            how: html
            open: true
    """

    _data_class = ReportHtmlData

    def prune(self, logger: tmt.log.Logger) -> None:
        """ Do not prune generated html report """

    def go(self, *, logger: Optional[tmt.log.Logger] = None) -> None:
        """ Process results """
        super().go(logger=logger)

        # Prepare the template
        environment = tmt.utils.templates.default_template_environment()

        # Explicitly enable the autoescape because it's disabled by default by TMT
        # (see /teemtee/tmt/issues/2873 for more info.
        environment.autoescape = select_autoescape(enabled_extensions=('html.j2'))

        if self.data.absolute_paths:
            def _linkable_path(path: str) -> str:
                return str(Path(path).absolute())

            environment.filters["linkable_path"] = _linkable_path
        else:
            # Links used in html should be relative to a workdir
            def _linkable_path(path: str) -> str:
                assert self.workdir is not None  # narrow type

                return str(Path(path).relative_to(self.workdir))

            environment.filters["linkable_path"] = _linkable_path

        if self.data.display_guest == 'always':
            display_guest = True

        elif self.data.display_guest == 'never':
            display_guest = False

        else:
            seen_guests = {
                result.guest.name
                for result in self.step.plan.execute.results()
                }

            display_guest = len(seen_guests) > 1

        # Write the report
        filename = Path('index.html')

        self.write(
            filename,
            data=tmt.utils.templates.render_template_file(
                HTML_TEMPLATE_PATH,
                environment,
                results=self.step.plan.execute.results(),
                base_dir=self.step.plan.execute.workdir,
                plan=self.step.plan,
                display_guest=display_guest))

        # Nothing more to do in dry mode
        if self.is_dry_run:
            return

        # Show output file path
        assert self.workdir is not None
        target = self.workdir / filename
        self.info("output", target, color='yellow')
        if not self.data.open:
            return

        # Open target in webbrowser
        try:
            if webbrowser.open(f"file://{target}", new=0):
                self.info(
                    'open', 'Successfully opened in the web browser.',
                    color='green')
                return
            self.fail("Failed to open the web browser.")
        except Exception as error:
            self.fail(f"Failed to open the web browser: {error}")

        raise tmt.utils.ReportError("Unable to open the web browser.")



import dataclasses
import os
import subprocess
import textwrap
from typing import Any, Optional, cast

import click
import jinja2

import tmt
import tmt.base
import tmt.log
import tmt.options
import tmt.steps
import tmt.steps.execute
import tmt.utils
from tmt.result import BaseResult, Result, ResultOutcome
from tmt.steps import safe_filename
from tmt.steps.execute import SCRIPTS, TEST_OUTPUT_FILENAME, TMT_REBOOT_SCRIPT, TestInvocation
from tmt.steps.provision import Guest
from tmt.utils import (
    Command,
    Environment,
    EnvVarValue,
    Path,
    ShellScript,
    Stopwatch,
    field,
    format_duration,
    format_timestamp,
    )

TEST_PIDFILE_FILENAME = 'tmt-test.pid'
TEST_PIDFILE_LOCK_FILENAME = f'{TEST_PIDFILE_FILENAME}.lock'

#: The default directory for storing test pid file.
TEST_PIDFILE_ROOT = Path('/var/tmp')  # noqa: S108 insecure usage of temporary dir


def effective_pidfile_root() -> Path:
    """
    Find out what the actual pidfile directory is.

    If ``TMT_TEST_PIDFILE_ROOT`` variable is set, it is used. Otherwise,
    :py:const:`TEST_PIDFILE_ROOT` is picked.
    """

    if 'TMT_TEST_PIDFILE_ROOT' in os.environ:
        return Path(os.environ['TMT_TEST_PIDFILE_ROOT'])

    return TEST_PIDFILE_ROOT


#: A template for the shell wrapper in which the test script is
#: saved.
#:
#: It is passed to :py:func:`tmt.utils.safe_filename`, but
#: includes also test name and serial number to make it unique
#: even among all test wrappers. See https://github.com/teemtee/tmt/issues/2997
#: for issue motivating the inclusion, it seems to be a good idea
#: to prevent accidental reuse in general.
TEST_WRAPPER_FILENAME_TEMPLATE = \
    'tmt-test-wrapper.sh-{{ INVOCATION.test.pathless_safe_name }}-{{ INVOCATION.test.serial_number }}'  # noqa: E501

# tmt test wrapper is getting complex. Besides honoring the timeout
# and interactivity request, it also must play nicely with reboots
# and `tmt-reboot`. The wrapper must present consistent info on what
# is the PID to kill from `tmt-reboot`, and where to save additional
# reboot info.
#
# For the duration of the test, the wrapper creates so-called "test
# pidfile". The pidfile contains test wrapper PID and path to the
# reboot-request file corresponding to the test being run. All actions
# against the pidfile must be taken while holding the pidfile lock,
# to serialize access between the wrapper and `tmt-reboot`. The file
# might be missing, that's allowed, but if it exists, it must contain
# correct info.
#
# Before quitting the wrapper, the pidfile is removed. There seems
# to be an apparent race condition: test quits -> `tmt-reboot` is
# called from a parallel session, grabs a pidfile lock, inspects
# pidfile, updates reboot-request, and sends signal to designed PID
# -> wrapper grabs the lock & removes the pidfile. This leaves us
# with `tmt-reboot` sending signal to non-existent PID - which is
# reported by `tmt-reboot`, "try again later" - and reboot-request
# file signaling reboot is needed *after the test is done*.
#
# This cannot be solved without the test being involved in the reboot,
# which does not seem like a viable option. The test must be restartable
# though, it may get restarted in this "weird" way. On the other hand,
# this is probably not a problem in real-life scenarios: tests that
# are to be interrupted by out-of-session reboot are expecting this
# action, and they do not finish on their own.
#
# The ssh client always allocates a tty, so test timeout handling
# works (#1387). Because the allocated tty is generally not suitable
# for test execution, the wrapper uses `|& cat` to emulate execution
# without a tty. In certain cases, where test execution with available
# tty is required (#2381), the tty can be kept on request with
# the `tty: true` test attribute.
#
# The wrapper script handles 3 execution modes for REMOTE_COMMAND:
#
# * In `tmt` interactive mode, stdin and stdout are unhandled, it is expected
#   user interacts with the executed command.
#
# * In non-interactive mode without a tty, stdin is fed with /dev/null (EOF)
#   and `|& cat` is used to simulate no tty available for script output.
#
# * In non-interactive mode with a tty, stdin is available to the tests
#   and simulation of tty not available for output is not run.
#
TEST_WRAPPER_TEMPLATE = jinja2.Template(textwrap.dedent("""
{% macro enter() %}
flock "$TMT_TEST_PIDFILE_LOCK" -c "echo '${test_pid} ${TMT_REBOOT_REQUEST}' > ${TMT_TEST_PIDFILE}" || exit 122
{%- endmacro %}

{% macro exit() %}
flock "$TMT_TEST_PIDFILE_LOCK" -c "rm -f ${TMT_TEST_PIDFILE}" || exit 123
{%- endmacro %}

[ ! -z "$TMT_DEBUG" ] && set -x

test_pid="$$";

mkdir -p "$(dirname $TMT_TEST_PIDFILE_LOCK)"

{% if INTERACTIVE %}
    {{ enter() }};
    {{ REMOTE_COMMAND }};
    _exit_code="$?";
    {{ exit() }};
{% elif TTY %}
    set -o pipefail;
    {{ enter() }};
    {{ REMOTE_COMMAND }} 2>&1;
    _exit_code="$?";
    {{ exit () }};
{% else %}
    set -o pipefail;
    {{ enter() }};
    {{ REMOTE_COMMAND }} </dev/null |& cat;
    _exit_code="$?";
    {{ exit () }};
{% endif %}
exit $_exit_code;
"""  # noqa: E501
))


class UpdatableMessage(tmt.utils.UpdatableMessage):
    """
    Updatable message suitable for plan progress reporting.

    Based on :py:class:`tmt.utils.UpdatableMessage`, simplifies
    reporting of plan progress, namely by extracting necessary setup
    parameters from the plugin.
    """

    def __init__(self, plugin: 'ExecuteInternal') -> None:
        super().__init__(
            key='progress',
            enabled=not plugin.verbosity_level and not plugin.data.no_progress_bar,
            indent_level=plugin._level(),
            key_color='cyan',
            clear_on_exit=True)

        self.plugin = plugin
        self.debug_level = plugin.debug_level

    # ignore[override]: the signature differs on purpose, we wish to input raw
    # values and let our `update()` construct the message.
    def update(self, progress: str, test_name: str) -> None:  # type: ignore[override]
        message = f'{test_name} [{progress}]'

        # With debug mode enabled, we do not really update a single line. Instead,
        # we shall emit each update as a distinct logging message, which would be
        # mixed into the debugging output.
        if self.debug_level:
            self.plugin.info(message)

        else:
            self._update_message_area(message)


@dataclasses.dataclass
class ExecuteInternalData(tmt.steps.execute.ExecuteStepData):
    script: list[ShellScript] = field(
        default_factory=list,
        option=('-s', '--script'),
        metavar='SCRIPT',
        multiple=True,
        help='Shell script to be executed as a test.',
        normalize=tmt.utils.normalize_shell_script_list,
        serialize=lambda scripts: [str(script) for script in scripts],
        unserialize=lambda serialized: [ShellScript(script) for script in serialized])
    interactive: bool = field(
        default=False,
        option=('-i', '--interactive'),
        is_flag=True,
        help="""
             Run tests in interactive mode, i.e. with input and output streams
             shared with tmt itself. This allows input to be passed to tests
             via stdin, e.g. responding to password prompts. Test output in this
             mode is not captured, and ``duration`` has no effect.
             """)
    no_progress_bar: bool = field(
        default=False,
        option='--no-progress-bar',
        is_flag=True,
        help='Disable interactive progress bar showing the current test.')

    # ignore[override] & cast: two base classes define to_spec(), with conflicting
    # formal types.
    def to_spec(self) -> dict[str, Any]:  # type: ignore[override]
        data = cast(dict[str, Any], super().to_spec())
        data['script'] = [str(script) for script in self.script]

        return data


@tmt.steps.provides_method('tmt')
class ExecuteInternal(tmt.steps.execute.ExecutePlugin[ExecuteInternalData]):
    """
    Use the internal tmt executor to execute tests

    The internal tmt executor runs tests on the guest one by one, shows
    testing progress and supports interactive debugging as well. Test
    result is based on the script exit code (for shell tests) or the
    results file (for beakerlib tests).
    """

    _data_class = ExecuteInternalData
    data: ExecuteInternalData

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)
        self._previous_progress_message = ""
        self.scripts = SCRIPTS

    def _test_environment(
            self,
            *,
            invocation: TestInvocation,
            extra_environment: Optional[Environment] = None,
            logger: tmt.log.Logger) -> Environment:
        """ Return test environment """

        extra_environment = extra_environment or Environment()

        environment = extra_environment.copy()
        environment.update(invocation.test.environment)
        assert self.parent is not None
        assert isinstance(self.parent, tmt.steps.execute.Execute)
        assert self.parent.plan.my_run is not None

        environment['TMT_TEST_PIDFILE'] = EnvVarValue(
            effective_pidfile_root() / TEST_PIDFILE_FILENAME)
        environment['TMT_TEST_PIDFILE_LOCK'] = EnvVarValue(
            effective_pidfile_root() / TEST_PIDFILE_LOCK_FILENAME)
        environment["TMT_TEST_NAME"] = EnvVarValue(invocation.test.name)
        environment["TMT_TEST_DATA"] = EnvVarValue(invocation.test_data_path)
        environment['TMT_TEST_SERIAL_NUMBER'] = EnvVarValue(str(invocation.test.serial_number))
        environment['TMT_TEST_ITERATION_ID'] = EnvVarValue(
            f"{self.parent.plan.my_run.unique_id}-{invocation.test.serial_number}")
        environment["TMT_TEST_METADATA"] = EnvVarValue(
            invocation.path / tmt.steps.execute.TEST_METADATA_FILENAME)
        environment["TMT_REBOOT_REQUEST"] = EnvVarValue(
            invocation.test_data_path / TMT_REBOOT_SCRIPT.created_file)
        # Set all supported reboot variables
        for reboot_variable in TMT_REBOOT_SCRIPT.related_variables:
            environment[reboot_variable] = EnvVarValue(str(invocation._reboot_count))
        environment['TMT_TEST_RESTART_COUNT'] = EnvVarValue(str(invocation._restart_count))

        # Add variables the framework wants to expose
        environment.update(
            invocation.test.test_framework.get_environment_variables(
                invocation, logger))

        return environment

    def _test_output_logger(
            self,
            key: str,
            value: Optional[str] = None,
            color: Optional[str] = None,
            shift: int = 2,
            level: int = 3,
            topic: Optional[tmt.log.Topic] = None) -> None:
        """ Custom logger for test output with shift 2 and level 3 defaults """
        self.verbose(key=key, value=value, color=color, shift=shift, level=level)

    def execute(
            self,
            *,
            invocation: TestInvocation,
            extra_environment: Optional[Environment] = None,
            logger: tmt.log.Logger) -> list[Result]:
        """ Run test on the guest """

        test, guest = invocation.test, invocation.guest

        logger.debug(f"Execute '{test.name}' as a '{test.framework}' test.")

        # Test will be executed in it's own directory, relative to the workdir
        assert self.discover.workdir is not None  # narrow type
        assert test.path is not None  # narrow type
        workdir = self.discover.workdir / test.path.unrooted()
        logger.debug(f"Use workdir '{workdir}'.", level=3)

        # Create data directory, prepare test environment
        environment = self._test_environment(
            invocation=invocation,
            extra_environment=extra_environment,
            logger=logger)

        # tmt wrapper filename *must* be "unique" - the plugin might be handling
        # the same `discover` phase for different guests at the same time, and
        # must keep them isolated. The wrapper script, while being prepared, is
        # a shared global state, and we must prevent race conditions.
        test_wrapper_filename = safe_filename(
            TEST_WRAPPER_FILENAME_TEMPLATE, self, guest, INVOCATION=invocation)
        test_wrapper_filepath = workdir / test_wrapper_filename

        logger.debug('test wrapper', test_wrapper_filepath)

        # Prepare the test command
        test_command = test.test_framework.get_test_command(invocation, logger)
        self.debug('Test script', test_command, level=3)

        # Prepare the wrapper, push to guest
        self.write(test_wrapper_filepath, str(test_command), 'w')
        test_wrapper_filepath.chmod(0o755)
        guest.push(
            source=test_wrapper_filepath,
            destination=test_wrapper_filepath,
            options=["-s", "-p", "--chmod=755"])

        # Create topology files
        topology = tmt.steps.Topology(self.step.plan.provision.guests())
        topology.guest = tmt.steps.GuestTopology(guest)

        environment.update(topology.push(
            dirpath=invocation.path,
            guest=guest,
            logger=logger))

        command: str
        if guest.become and not guest.facts.is_superuser:
            command = f'sudo -E ./{test_wrapper_filename}'
        else:
            command = f'./{test_wrapper_filename}'
        # Prepare the actual remote command
        remote_command = ShellScript(TEST_WRAPPER_TEMPLATE.render(
            INTERACTIVE=self.data.interactive,
            TTY=test.tty,
            REMOTE_COMMAND=ShellScript(command)
            ).strip())

        def _test_output_logger(
                key: str,
                value: Optional[str] = None,
                color: Optional[str] = None,
                shift: int = 2,
                level: int = 3,
                topic: Optional[tmt.log.Topic] = None) -> None:
            logger.verbose(
                key=key,
                value=value,
                color=color,
                shift=shift,
                level=level,
                topic=topic)

        def _save_process(
                command: Command,
                process: subprocess.Popen[bytes],
                logger: tmt.log.Logger) -> None:
            with invocation.process_lock:
                invocation.process = process

        # TODO: do we want timestamps? Yes, we do, leaving that for refactoring later,
        # to use some reusable decorator.
        invocation.check_results = self.run_checks_before_test(
            invocation=invocation,
            environment=environment,
            logger=logger
            )

        # Execute the test, save the output and return code
        with Stopwatch() as timer:
            invocation.start_time = format_timestamp(timer.start_time)

            timeout: Optional[int]

            if self.data.interactive:
                if test.duration:
                    logger.warning(
                        'Ignoring requested duration, not supported in interactive mode.')

                timeout = None

            else:
                timeout = tmt.utils.duration_to_seconds(
                    test.duration, tmt.base.DEFAULT_TEST_DURATION_L1)

            try:
                output = guest.execute(
                    remote_command,
                    cwd=workdir,
                    env=environment,
                    join=True,
                    interactive=self.data.interactive,
                    tty=test.tty,
                    log=_test_output_logger,
                    timeout=timeout,
                    on_process_start=_save_process,
                    test_session=True,
                    friendly_command=str(test.test))
                invocation.return_code = 0
                stdout = output.stdout
            except tmt.utils.RunError as error:
                stdout = error.stdout

                invocation.return_code = error.returncode
                if invocation.return_code == tmt.utils.ProcessExitCodes.TIMEOUT:
                    logger.debug(f"Test duration '{test.duration}' exceeded.")

                elif tmt.utils.ProcessExitCodes.is_pidfile(invocation.return_code):
                    logger.warning('Test failed to manage its pidfile.')

        with invocation.process_lock:
            invocation.process = None

        invocation.end_time = format_timestamp(timer.end_time)
        invocation.real_duration = format_duration(timer.duration)

        # Save the captured output. Do not let the follow-up pulls
        # overwrite it.
        self.write(
            invocation.path / TEST_OUTPUT_FILENAME,
            stdout or '', mode='a', level=3)

        # Fetch #1: we need logs and everything the test produced so we could
        # collect its results.
        if invocation.is_guest_healthy:
            guest.pull(
                source=invocation.path,
                extend_options=[
                    *test.test_framework.get_pull_options(invocation, logger),
                    '--exclude', str(invocation.path / TEST_OUTPUT_FILENAME)
                    ])
            # Fetch plan data content as well in order to prevent
            # losing logs if the guest becomes later unresponsive.
            guest.pull(source=self.step.plan.data_directory)

        # Run after-test checks before extracting results
        invocation.check_results += self.run_checks_after_test(
            invocation=invocation,
            environment=environment,
            logger=logger
            )

        # Extract test results and store them in the invocation. Note
        # that these results will be overwritten with a fresh set of
        # results after a successful reboot in the middle of a test.
        invocation.results = self.extract_results(invocation, logger)

        if invocation.is_guest_healthy:
            # Fetch #2: after-test checks might have produced remote files as well,
            # we need to fetch them too.
            guest.pull(
                source=invocation.path,
                extend_options=[
                    *test.test_framework.get_pull_options(invocation, logger),
                    '--exclude', str(invocation.path / TEST_OUTPUT_FILENAME)
                    ])
            # Fetch plan data content as well in order to prevent
            # losing logs if the guest becomes later unresponsive.
            guest.pull(source=self.step.plan.data_directory)

        # Attach check results to every test result. There might be more than one,
        # and it's hard to pick the main one, who knows what custom results might
        # cover, so let's make sure every single result can lead to check results
        # related to its lifetime.
        for result in invocation.results:
            result.check = invocation.check_results

        return invocation.results

    def go(
            self,
            *,
            guest: 'Guest',
            environment: Optional[tmt.utils.Environment] = None,
            logger: tmt.log.Logger) -> None:
        """ Execute available tests """
        super().go(guest=guest, environment=environment, logger=logger)

        # Nothing to do in dry mode
        if self.is_dry_run:
            self._results = []
            return

        self._run_tests(guest=guest, extra_environment=environment, logger=logger)

    def _run_tests(
            self,
            *,
            guest: Guest,
            extra_environment: Optional[Environment] = None,
            logger: tmt.log.Logger) -> None:
        """ Execute tests on provided guest """

        # Prepare tests and helper scripts, check options
        test_invocations = self.prepare_tests(guest, logger)

        # Prepare scripts, except localhost guest
        if not guest.localhost:
            self.prepare_scripts(guest)

        # Push workdir to guest and execute tests
        guest.push()
        # We cannot use enumerate here due to continue in the code
        index = 0

        with UpdatableMessage(self) as progress_bar:
            while index < len(test_invocations):
                invocation = test_invocations[index]

                test = invocation.test

                progress = f"{index + 1}/{len(test_invocations)}"
                progress_bar.update(progress, test.name)
                logger.verbose(
                    'test', test.summary or test.name, color='cyan', shift=1, level=2)

                self.execute(
                    invocation=invocation,
                    extra_environment=extra_environment,
                    logger=logger)

                assert invocation.real_duration is not None  # narrow type
                duration = click.style(invocation.real_duration, fg='cyan')
                shift = 1 if self.verbosity_level < 2 else 2

                # Handle test restart. May include guest reboot too.
                if invocation.restart_requested:
                    # Output before the restart
                    logger.verbose(f"{duration} {test.name} [{progress}]", shift=shift)

                    try:
                        if invocation.handle_restart():
                            continue

                    except tmt.utils.RebootTimeoutError:
                        for result in invocation.results:
                            result.result = ResultOutcome.ERROR
                            result.note = 'reboot timeout'

                    else:
                        for result in invocation.results:
                            result.result = ResultOutcome.ERROR
                            result.note = 'crashed too many times'

                # Handle reboot, abort, exit-first
                if invocation.reboot_requested:
                    # Output before the reboot
                    logger.verbose(
                        f"{duration} {test.name} [{progress}]", shift=shift)
                    try:
                        if invocation.handle_reboot():
                            continue
                    except tmt.utils.RebootTimeoutError:
                        for result in invocation.results:
                            result.result = ResultOutcome.ERROR
                            result.note = 'reboot timeout'

                if invocation.abort_requested:
                    for result in invocation.results:
                        # In case of aborted all results in list will be aborted
                        result.note = 'aborted'

                self._results.extend(invocation.results)

                # If test duration information is missing, print 8 spaces to keep indentation
                def _format_duration(result: BaseResult) -> str:
                    return click.style(result.duration, fg='cyan') if result.duration else 8 * ' '

                for result in invocation.results:
                    logger.verbose(
                        f"{_format_duration(result)} {result.show()} [{progress}]",
                        shift=shift)

                    for check_result in result.check:
                        # Indent the check one extra level, to make it clear it belongs to
                        # a parent test.
                        logger.verbose(
                            f'{_format_duration(check_result)} '
                            f'{" " * tmt.utils.INDENT}'
                            f'{check_result.show()} '
                            f'({check_result.event.value} check)',
                            shift=shift)

                if invocation.abort_requested or (
                    self.data.exit_first and any(
                        result.result not in (
                            ResultOutcome.PASS,
                            ResultOutcome.INFO) for result in invocation.results)):
                    # Clear the progress bar before outputting
                    progress_bar.clear()
                    what_happened = "aborted" if invocation.abort_requested else "failed"
                    self.warn(
                        f'Test {test.name} {what_happened}, stopping execution.')
                    break
                index += 1

                # Log into the guest after each executed test if "login
                # --test" option is provided
                if self._login_after_test:
                    assert test.path is not None  # narrow type

                    if self.discover.workdir is None:
                        cwd = test.path.unrooted()
                    else:
                        cwd = self.discover.workdir / test.path.unrooted()
                    self._login_after_test.after_test(
                        result,
                        cwd=cwd,
                        env=self._test_environment(
                            invocation=invocation,
                            extra_environment=extra_environment,
                            logger=logger),
                        )

        # Pull artifacts created in the plan data directory
        self.debug("Pull the plan data directory.", level=2)
        guest.pull(source=self.step.plan.data_directory)

    def results(self) -> list[Result]:
        """ Return test results """
        return self._results

    def essential_requires(self) -> list[tmt.base.Dependency]:
        """
        Collect all essential requirements of the plugin.

        Essential requirements of a plugin are necessary for the plugin to
        perform its basic functionality.

        :returns: a list of requirements.
        """

        return [
            tmt.base.DependencySimple('/usr/bin/flock')
            ]



import re
from typing import TYPE_CHECKING, Optional

import tmt.base
import tmt.log
import tmt.result
import tmt.steps.execute
import tmt.utils
from tmt.frameworks import TestFramework, provides_framework
from tmt.result import ResultOutcome
from tmt.utils import Environment, EnvVarValue, Path

if TYPE_CHECKING:
    from tmt.base import DependencySimple, Test
    from tmt.steps.execute import TestInvocation


@provides_framework('beakerlib')
class Beakerlib(TestFramework):
    @classmethod
    def get_requirements(
            cls,
            test: 'Test',
            logger: tmt.log.Logger) -> list['DependencySimple']:
        # Avoiding circular imports: `Test.test_framework` requires `tmt.frameworks`.
        from tmt.base import DependencySimple

        return [DependencySimple('beakerlib')]

    @classmethod
    def get_environment_variables(
            cls,
            invocation: 'TestInvocation',
            logger: tmt.log.Logger) -> tmt.utils.Environment:

        return Environment({
            'BEAKERLIB_DIR': EnvVarValue(invocation.path),
            'BEAKERLIB_COMMAND_SUBMIT_LOG': EnvVarValue(
                invocation.guest.scripts_path /
                tmt.steps.execute.TMT_FILE_SUBMIT_SCRIPT.source_filename),

            # The command in this variable gets called with every `rlPhaseEnd` call in beakerlib.
            'BEAKERLIB_COMMAND_REPORT_RESULT': EnvVarValue(
                invocation.guest.scripts_path /
                tmt.steps.execute.TMT_REPORT_RESULT_SCRIPT.source_filename),

            # This variables must be set explicitly, otherwise the beakerlib `rlPhaseEnd` macro
            # will not call the the command in `BEAKERLIB_COMMAND_REPORT_RESULT`.
            # - https://github.com/beakerlib/beakerlib/blob/cfa801fb175fef1e47b8552d6cf6efcb51df7227/src/testing.sh#L1074
            'TESTID': EnvVarValue(str(invocation.test.serial_number)),
            })

    @classmethod
    def get_pull_options(
            cls,
            invocation: 'TestInvocation',
            logger: tmt.log.Logger) -> list[str]:
        return [
            '--exclude',
            str(invocation.path / "backup*")
            ]

    @classmethod
    def extract_results(
            cls,
            invocation: 'TestInvocation',
            results: list[tmt.result.Result],
            logger: tmt.log.Logger) -> list[tmt.result.Result]:
        """ Check result of a beakerlib test """

        # The outcome of a main tmt result must be never modified based on subresults outcomes.
        # The main result outcome will be always set to outcome reported by a beakerlib. The
        # subresults are there just to provide more detail.
        subresults = [result.to_subresult() for result in results]

        # Initialize data, prepare log paths
        note: Optional[str] = None
        log: list[Path] = []
        for filename in [tmt.steps.execute.TEST_OUTPUT_FILENAME, 'journal.txt']:
            if (invocation.path / filename).is_file():
                log.append(invocation.relative_path / filename)

        # Check beakerlib log for the result
        beakerlib_results_filepath = invocation.path / 'TestResults'

        try:
            beakerlib_results = invocation.phase.read(beakerlib_results_filepath, level=3)
        except tmt.utils.FileError:
            logger.debug(f"Unable to read '{beakerlib_results_filepath}'.", level=3)
            note = 'beakerlib: TestResults FileError'

            return [tmt.Result.from_test_invocation(
                invocation=invocation,
                result=ResultOutcome.ERROR,
                note=note,
                log=log,
                subresult=subresults)]

        search_result = re.search('TESTRESULT_RESULT_STRING=(.*)', beakerlib_results)
        # States are: started, incomplete and complete
        # FIXME In quotes until beakerlib/beakerlib/pull/92 is merged
        search_state = re.search(r'TESTRESULT_STATE="?(\w+)"?', beakerlib_results)

        if search_result is None or search_state is None:
            # Same outcome but make it easier to debug
            if search_result is None:
                missing_piece = 'TESTRESULT_RESULT_STRING='
                hint = ''
            else:
                missing_piece = 'TESTRESULT_STATE='
                hint = ', possibly outdated beakerlib (requires 1.23+)'
            logger.debug(
                f"No '{missing_piece}' found in '{beakerlib_results_filepath}'{hint}.",
                level=3)
            note = 'beakerlib: Result/State missing'
            return [tmt.Result.from_test_invocation(
                invocation=invocation,
                result=ResultOutcome.ERROR,
                note=note,
                log=log,
                subresult=subresults)]

        result = search_result.group(1)
        state = search_state.group(1)

        # Check if it was killed by timeout (set by tmt executor)
        actual_result = ResultOutcome.ERROR
        if invocation.return_code == tmt.utils.ProcessExitCodes.TIMEOUT:
            note = 'timeout'
            invocation.phase.timeout_hint(invocation)

        elif tmt.utils.ProcessExitCodes.is_pidfile(invocation.return_code):
            note = 'pidfile locking'

        # Test results should be in complete state
        elif state != 'complete':
            note = f"beakerlib: State '{state}'"
        # Finally we have a valid result
        else:
            actual_result = ResultOutcome.from_spec(result.lower())

        return [tmt.Result.from_test_invocation(
            invocation=invocation,
            result=actual_result,
            note=note,
            log=log,
            subresult=subresults)]
